from __future__ import annotations

import abc
import ast
from typing import Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.symbol import ISymbol
from boa3.model.type.itype import IType
from boa3.model.variable import Variable


class IInternalMethod(IBuiltinMethod, abc.ABC):

    def __init__(self, identifier: str, args: Dict[str, Variable] = None,
                 defaults: List[ast.AST] = None, return_type: IType = None,
                 vararg: Optional[Tuple[str, Variable]] = None):
        super().__init__(identifier, args, defaults, return_type, vararg)

    @classmethod
    @abc.abstractmethod
    def instance(cls) -> IInternalMethod:
        pass

    @classmethod
    @abc.abstractmethod
    def is_valid_deploy_method(cls, symbol: ISymbol) -> bool:
        return False



from typing import Dict, List, Tuple

from boa3.model.builtin.builtinproperty import IBuiltinProperty
from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.builtin.interop.iterator.iteratortype import IteratorType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class GetIteratorValue(InteropMethod):
    def __init__(self, iterator: IteratorType):
        syscall = 'System.Iterator.Value'
        identifier = '-get_iterator_value'
        args: Dict[str, Variable] = {'self': Variable(iterator)}

        result_type = iterator.item_type
        from boa3.model.type.collection.mapping.mappingtype import MappingType
        if isinstance(iterator._origin_collection, MappingType):
            from boa3.model.type.type import Type
            result_type = Type.tuple.build((iterator.key_type, iterator.item_type))
        super().__init__(identifier, syscall, args, return_type=result_type)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def self_arg(self) -> Variable:
        return self.args['self']

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.neo.vm.type.StackItem import StackItemType
        from boa3.neo.vm.type.Integer import Integer
        return super()._opcode + [
            (Opcode.DUP, b''),
            (Opcode.ISTYPE, StackItemType.Struct),
            (Opcode.JMPIFNOT, Integer(3).to_byte_array()),
            (Opcode.CONVERT, StackItemType.Array),
        ]


class IteratorValueProperty(IBuiltinProperty):
    def __init__(self, iterator: IteratorType):
        identifier = 'iterator_value'
        getter = GetIteratorValue(iterator)
        super().__init__(identifier, getter)



import ast
from typing import Any, Dict, List, Optional

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.variable import Variable


class StrMethod(IBuiltinMethod):

    def __init__(self, args: Dict[str, Variable] = None, defaults: List[ast.AST] = None):
        from boa3.model.type.type import Type
        identifier = 'str'
        super().__init__(identifier, args, defaults=defaults, return_type=Type.str)

    @property
    def _arg_value(self) -> Variable:
        return self.args['object']

    @property
    def identifier(self) -> str:
        from boa3.model.type.type import Type

        if self._arg_value.type is Type.bool:
            return '-{0}_{1}'.format(self._identifier, Type.bool.identifier)

        if self._arg_value.type is Type.int:
            return '-{0}_{1}'.format(self._identifier, Type.int.identifier)

        return self._identifier

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if not isinstance(value, list):
            value = [value]

        from boa3.model.type.type import Type
        from boa3.model.builtin.builtin import Builtin
        if len(value) > 0:

            if Type.bool.is_type_of(value[0]):
                return Builtin.StrBool

            if Type.int.is_type_of(value[0]):
                return Builtin.StrInt

        return Builtin.StrByteString



from typing import Any

from boa3.model.type.collection.sequence.sequencetype import SequenceType
from boa3.model.type.itype import IType


class RangeType(SequenceType):
    """
    A class used to represent Python range type
    """

    def __init__(self, values_type: IType):
        identifier = 'range'
        values_type = self.filter_types(values_type)
        super().__init__(identifier, values_type)

    @property
    def default_value(self) -> Any:
        return range(0)

    @property
    def identifier(self) -> str:
        return self._identifier

    def is_valid_key(self, key_type: IType) -> bool:
        return key_type == self.valid_key

    @property
    def valid_key(self) -> IType:
        from boa3.model.type.type import Type
        return Type.int

    @classmethod
    def build(cls, value: Any) -> IType:
        if cls._is_type_of(value):
            from boa3.model.type.type import Type
            return cls(Type.int)

    @classmethod
    def _is_type_of(cls, value: Any):
        return type(value) is range or isinstance(value, RangeType)

    @property
    def can_reassign_values(self) -> bool:
        return False

    def __eq__(self, other) -> bool:
        if type(self) != type(other):
            return False
        return self.value_type == other.value_type

    def __hash__(self):
        return hash(self.identifier)



import ast
from abc import ABC
from typing import Any, Dict, List

from boa3.model.builtin.builtincallable import IBuiltinCallable
from boa3.model.decorator import IDecorator
from boa3.model.expression import IExpression
from boa3.model.type.itype import IType
from boa3.model.variable import Variable


class IBuiltinDecorator(IBuiltinCallable, IDecorator, ABC):
    def __init__(self, identifier: str, args: Dict[str, Variable] = None,
                 defaults: List[ast.AST] = None, return_type: IType = None):
        super().__init__(identifier, args, defaults, return_type)

    def validate_parameters(self, *params: IExpression) -> bool:
        return len(params) == len(self.args)

    def validate_values(self, *params: Any) -> List[Any]:
        return []



from typing import List

from boa3.builtin.type import UInt160
from boa3.neo3.network.payloads.verification import WitnessConditionType, WitnessRuleAction, WitnessScope


class Signer:
    """
    Represents a signer.

    :ivar account:
    :vartype account: UInt160
    :ivar scopes:
    :vartype scopes: WitnessScope
    :ivar allowed_contracts:
    :vartype allowed_contracts: List[UInt160]
    :ivar allowed_groups:
    :vartype allowed_groups: List[UInt160]
    :ivar rules:
    :vartype rules: List[WitnessRule]
    """

    def __init__(self):
        self.account: UInt160 = UInt160()
        self.scopes: WitnessScope = WitnessScope.NONE
        self.allowed_contracts: List[UInt160] = []
        self.allowed_groups: List[UInt160] = []
        self.rules: List[WitnessRule] = []


class WitnessRule:
    """
    Represents a witness rule.

    :ivar action:
    :vartype action: WitnessRuleAction
    :ivar condition:
    :vartype condition: WitnessCondition
    """

    def __init__(self):
        self.action: WitnessRuleAction = WitnessRuleAction.DENY
        self.condition: WitnessCondition = WitnessCondition()


class WitnessCondition:
    """
    Represents a witness condition.

    :ivar type:
    :vartype type: WitnessConditionType
    """

    def __init__(self):
        self.type: WitnessConditionType = WitnessConditionType.BOOLEAN



import os
from typing import Optional

from boa3.neo import to_hex_str
from boa3.neo.vm.type.StackItem import StackItemType


class TestContract:
    def __init__(self, file_path: str, manifest_path: str):
        self._nef_path: str = file_path
        self._manifest_path: str = manifest_path
        self._manifest: dict = self._read_manifest()
        self._script_hash: Optional[bytes] = self._get_script_hash()

    def _get_script_hash(self) -> bytes:
        file_path = self._nef_path
        if file_path.endswith('.nef') and os.path.isfile(file_path):
            with open(file_path, mode='rb') as nef:
                file = nef.read()

            from boa3.neo import to_script_hash
            from boa3.neo.contracts.neffile import NefFile
            from boa3.neo.vm.opcode.Opcode import Opcode
            from boa3.neo.vm.opcode import OpcodeHelper

            from boa3.neo3.core.types import UInt160
            push_sender, sender = OpcodeHelper.get_pushdata_and_data(
                UInt160.from_string("0x469cc2649dd10365ffd1e5e25af67b22042fa59c").to_array()  # mocked sender
            )
            push_checksum, checksum = OpcodeHelper.get_push_and_data(NefFile.deserialize(file).checksum)
            push_name, contract_name = OpcodeHelper.get_pushdata_and_data(self._manifest['name'])

            script = (Opcode.ABORT
                      + push_sender + sender
                      + push_checksum + checksum
                      + push_name + contract_name)
            return to_script_hash(script)

    def _read_manifest(self) -> dict:
        try:
            with open(self._manifest_path) as manifest_output:
                import json
                manifest = json.loads(manifest_output.read())
        except BaseException:
            manifest = {}

        return manifest

    @property
    def path(self) -> str:
        return self._nef_path

    @property
    def script_hash(self) -> Optional[bytes]:
        return self._script_hash

    def __str__(self) -> str:
        return f'[{to_hex_str(self._script_hash)}] {self._nef_path}'

    def __repr__(self) -> str:
        return str(self)

    def is_valid_identifier(self, item: str) -> bool:
        # valid identifiers are nef path or a representation of the script hash
        return (item == self._nef_path
                or item == self._script_hash
                or item == to_hex_str(self._script_hash))

    def get_method_return_type(self, called_method: str) -> StackItemType:
        cur_json = self._manifest
        return_type = StackItemType.Any

        if 'abi' in cur_json:
            cur_json = cur_json['abi']
            if 'methods' in cur_json:
                cur_json = cur_json['methods']
                method_json = next((node for node in cur_json
                                    if 'name' in node and node['name'] == called_method),
                                   {})
                if 'returntype' in method_json:
                    return_type = StackItemType.get_stack_item_type(method_json['returntype'])

        return return_type



from boa3 import constants
from boa3.neo import to_script_hash
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestNEP17Template(BoaTest):
    default_folder: str = 'examples'

    OWNER_SCRIPT_HASH = bytes(20)
    OTHER_ACCOUNT_1 = to_script_hash(b'NiNmXL8FjEUEs1nfX9uHFBNaenxDHJtmuB')
    OTHER_ACCOUNT_2 = bytes(range(20))

    def test_nep17_compile(self):
        path = self.get_contract_path('nep17.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('supportedstandards', manifest)
        self.assertIsInstance(manifest['supportedstandards'], list)
        self.assertGreater(len(manifest['supportedstandards']), 0)
        self.assertIn('NEP-17', manifest['supportedstandards'])

    def test_nep17_symbol(self):
        path = self.get_contract_path('nep17.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'symbol')
        self.assertEqual('NEP17', result)

    def test_nep17_decimals(self):
        path = self.get_contract_path('nep17.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'decimals')
        self.assertEqual(8, result)

    def test_nep17_total_supply(self):
        total_supply = 10_000_000 * 10 ** 8

        path = self.get_contract_path('nep17.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'totalSupply')
        self.assertEqual(total_supply, result)

    def test_nep17_total_balance_of(self):
        total_supply = 10_000_000 * 10 ** 8

        path = self.get_contract_path('nep17.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        self.assertEqual(total_supply, result)

        # should fail when the script length is not 20
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'balanceOf', bytes(10))
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'balanceOf', bytes(30))

    def test_nep17_total_transfer(self):
        transferred_amount = 10 * 10 ** 8  # 10 tokens

        path = self.get_contract_path('nep17.py')
        engine = TestEngine()

        # should fail if the sender doesn't sign
        result = self.run_smart_contract(engine, path, 'transfer',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, transferred_amount, None)
        self.assertEqual(False, result)

        # should fail if the sender doesn't have enough balance
        result = self.run_smart_contract(engine, path, 'transfer',
                                         self.OTHER_ACCOUNT_1, self.OWNER_SCRIPT_HASH, transferred_amount, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(False, result)

        # should fail when any of the scripts' length is not 20
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'transfer',
                                    self.OWNER_SCRIPT_HASH, bytes(10), transferred_amount, "")
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'transfer',
                                    bytes(10), self.OTHER_ACCOUNT_1, transferred_amount, "")

        # should fail when the amount is less than 0
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'transfer',
                                    self.OTHER_ACCOUNT_1, self.OWNER_SCRIPT_HASH, -10, None)

        # fire the transfer event when transferring to yourself
        balance_before = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        result = self.run_smart_contract(engine, path, 'transfer',
                                         self.OWNER_SCRIPT_HASH, self.OWNER_SCRIPT_HASH, transferred_amount, None,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)
        transfer_events = engine.get_events('Transfer')
        self.assertEqual(2, len(transfer_events))
        self.assertEqual(3, len(transfer_events[1].arguments))

        sender, receiver, amount = transfer_events[1].arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(self.OWNER_SCRIPT_HASH, sender)
        self.assertEqual(self.OWNER_SCRIPT_HASH, receiver)
        self.assertEqual(transferred_amount, amount)

        # transferring to yourself doesn't change the balance
        balance_after = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        self.assertEqual(balance_before, balance_after)

        # does fire the transfer event
        balance_sender_before = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        balance_receiver_before = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        result = self.run_smart_contract(engine, path, 'transfer',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, transferred_amount, None,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)
        transfer_events = engine.get_events('Transfer')
        self.assertEqual(3, len(transfer_events))
        self.assertEqual(3, len(transfer_events[2].arguments))

        sender, receiver, amount = transfer_events[2].arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(self.OWNER_SCRIPT_HASH, sender)
        self.assertEqual(self.OTHER_ACCOUNT_1, receiver)
        self.assertEqual(transferred_amount, amount)

        # transferring to someone other than yourself does change the balance
        balance_sender_after = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        balance_receiver_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        self.assertEqual(balance_sender_before - transferred_amount, balance_sender_after)
        self.assertEqual(balance_receiver_before + transferred_amount, balance_receiver_after)

    def test_nep17_on_nep17_payment(self):
        transferred_amount = 10 * 10 ** 8  # 10 tokens

        path = self.get_contract_path('nep17.py')
        engine = TestEngine()

        self.run_smart_contract(engine, path, 'symbol')
        nep17_address = engine.executed_script_hash.to_array()

        engine.reset_engine()
        engine.add_contract(path.replace('.py', '.nef'))

        engine.add_neo(self.OTHER_ACCOUNT_1, transferred_amount)
        engine.add_gas(self.OTHER_ACCOUNT_1, transferred_amount)

        # transferring NEO to the smart contract
        # saving the balance before the transfer to be able to compare after it
        neo_balance_sender_before = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf',
                                                            self.OTHER_ACCOUNT_1)
        neo_balance_nep17_before = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf', nep17_address)
        nep17_balance_sender_before = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)

        result = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'transfer',
                                         self.OTHER_ACCOUNT_1, nep17_address, transferred_amount, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)
        transfer_events = engine.get_events('Transfer')
        self.assertEqual(4, len(transfer_events))
        neo_transfer_event = transfer_events[1]
        transfer_event = transfer_events[2]
        self.assertEqual(3, len(neo_transfer_event.arguments))
        self.assertEqual(3, len(transfer_event.arguments))

        # this is the event NEO emitted
        sender, receiver, amount = neo_transfer_event.arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(self.OTHER_ACCOUNT_1, sender)
        self.assertEqual(nep17_address, receiver)
        self.assertEqual(transferred_amount, amount)

        # this is the event NEP17 emitted
        sender, receiver, amount = transfer_event.arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(None, sender)
        self.assertEqual(self.OTHER_ACCOUNT_1, receiver)
        # transferred_amount is multiplied by 10, because this smart contract is minting the NEO received
        self.assertEqual(transferred_amount * 10, amount)

        # saving the balance after the transfer to compare it with the previous data
        neo_balance_sender_after = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf',
                                                           self.OTHER_ACCOUNT_1)
        neo_balance_nep17_after = self.run_smart_contract(engine, constants.NEO_SCRIPT, 'balanceOf', nep17_address)
        nep17_balance_sender_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)

        self.assertEqual(neo_balance_sender_before - transferred_amount, neo_balance_sender_after)
        self.assertEqual(neo_balance_nep17_before + transferred_amount, neo_balance_nep17_after)
        # transferred_amount is multiplied by 10, because this smart contract is minting the NEO received
        self.assertEqual(nep17_balance_sender_before + transferred_amount * 10, nep17_balance_sender_after)

        # transferring GAS to the smart contract
        # saving the balance before the transfer to be able to compare after it
        gas_balance_sender_before = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'balanceOf',
                                                            self.OTHER_ACCOUNT_1)
        gas_balance_nep17_before = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'balanceOf', nep17_address)
        nep17_balance_sender_before = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)

        result = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'transfer',
                                         self.OTHER_ACCOUNT_1, nep17_address, transferred_amount, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)
        transfer_events = engine.get_events('Transfer')
        self.assertEqual(6, len(transfer_events))
        gas_transfer_event = transfer_events[4]
        transfer_event = transfer_events[5]
        self.assertEqual(3, len(gas_transfer_event.arguments))
        self.assertEqual(3, len(transfer_event.arguments))

        # this is the event GAS emitted
        sender, receiver, amount = gas_transfer_event.arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(self.OTHER_ACCOUNT_1, sender)
        self.assertEqual(nep17_address, receiver)
        self.assertEqual(transferred_amount, amount)

        # this is the event NEP17 emitted
        sender, receiver, amount = transfer_event.arguments
        if isinstance(sender, str):
            sender = String(sender).to_bytes()
        if isinstance(receiver, str):
            receiver = String(receiver).to_bytes()
        self.assertEqual(None, sender)
        self.assertEqual(self.OTHER_ACCOUNT_1, receiver)
        # transferred_amount is multiplied by 2, because this smart contract is minting the GAS received
        self.assertEqual(transferred_amount * 2, amount)

        # saving the balance after the transfer to compare it with the previous data
        gas_balance_sender_after = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'balanceOf',
                                                           self.OTHER_ACCOUNT_1)
        gas_balance_nep17_after = self.run_smart_contract(engine, constants.GAS_SCRIPT, 'balanceOf', nep17_address)
        nep17_balance_sender_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)

        self.assertEqual(gas_balance_sender_before - transferred_amount, gas_balance_sender_after)
        self.assertEqual(gas_balance_nep17_before + transferred_amount, gas_balance_nep17_after)
        # transferred_amount is multiplied by 2, because this smart contract is minting the GAS received
        self.assertEqual(nep17_balance_sender_before + transferred_amount * 2, nep17_balance_sender_after)

        # trying to call onNEP17Transfer() will result in an abort if the one calling it is not NEO or GAS contracts
        with self.assertRaisesRegex(TestExecutionException, f'{self.CANT_FIND_METHOD_MSG_PREFIX} : onNEP17Transfer'):
            self.run_smart_contract(engine, path, 'onNEP17Transfer',
                                    self.OWNER_SCRIPT_HASH, transferred_amount, None)

    def test_nep17_verify(self):
        path = self.get_contract_path('nep17.py')
        engine = TestEngine()

        # should fail without signature
        result = self.run_smart_contract(engine, path, 'verify')
        self.assertEqual(False, result)

        # should fail if not signed by the owner
        result = self.run_smart_contract(engine, path, 'verify',
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'verify',
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)



from boa3 import constants
from boa3.constants import IMPORT_WILDCARD
from boa3.exception import CompilerError, CompilerWarning
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo3.core.types import UInt160
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.contract.neomanifeststruct import NeoManifestStruct
from boa3_test.tests.test_classes.testengine import TestEngine


class TestMetadata(BoaTest):
    default_folder: str = 'test_sc/metadata_test'

    def test_metadata_info_method(self):
        expected_output = (
            Opcode.PUSH5  # return 5
            + Opcode.RET
        )

        path = self.get_contract_path('MetadataInfoMethod.py')
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        # test extra field
        self.assertIn('extra', manifest)
        self.assertIsNone(manifest['extra'])

    def test_metadata_info_method_mismatched_type(self):
        path = self.get_contract_path('MetadataInfoMethodMismatchedReturn.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_metadata_info_method_no_return(self):
        path = self.get_contract_path('MetadataInfoMethodNoReturn.py')
        self.assertCompilerLogs(CompilerError.MissingReturnStatement, path)

    def test_metadata_info_multiple_method(self):
        expected_output = (
            Opcode.PUSH5  # return 5
            + Opcode.RET
        )

        path = self.get_contract_path('MetadataInfoMultipleMethod.py')
        self.assertCompilerLogs(CompilerWarning.RedeclaredSymbol, path)

        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        self.assertIn('extra', manifest)
        self.assertIsNotNone(manifest['extra'])
        self.assertIn('Description', manifest['extra'])
        self.assertEqual('func1', manifest['extra']['Description'])

    def test_metadata_method_with_args(self):
        path = self.get_contract_path('MetadataMethodWithArgs.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_metadata_method_called_by_user_method(self):
        path = self.get_contract_path('MetadataMethodCalledByUserMethod.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_metadata_object_call_user_method(self):
        path = self.get_contract_path('MetadataObjectCallUserMethod.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_metadata_object_type_user_method(self):
        path = self.get_contract_path('MetadataObjectTypeUserMethod.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_metadata_info_author(self):
        expected_output = (
            Opcode.PUSH5  # return 5
            + Opcode.RET
        )

        path = self.get_contract_path('MetadataInfoAuthor.py')
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        self.assertIn('extra', manifest)
        self.assertIsNotNone(manifest['extra'])
        self.assertIn('Author', manifest['extra'])
        self.assertEqual('Test', manifest['extra']['Author'])

    def test_metadata_info_author_mismatched_type(self):
        path = self.get_contract_path('MetadataInfoAuthorMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_metadata_info_email(self):
        expected_output = (
            Opcode.PUSH5  # return 5
            + Opcode.RET
        )

        path = self.get_contract_path('MetadataInfoEmail.py')
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        self.assertIn('extra', manifest)
        self.assertIsNotNone(manifest['extra'])
        self.assertIn('Email', manifest['extra'])
        self.assertEqual('test@test.com', manifest['extra']['Email'])

    def test_metadata_info_email_mismatched_type(self):
        path = self.get_contract_path('MetadataInfoEmailMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_metadata_info_description(self):
        expected_output = (
            Opcode.PUSH5  # return 5
            + Opcode.RET
        )

        path = self.get_contract_path('MetadataInfoDescription.py')
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        self.assertIn('extra', manifest)
        self.assertIsNotNone(manifest['extra'])
        self.assertIn('Description', manifest['extra'])
        self.assertEqual('This is an example', manifest['extra']['Description'])

    def test_metadata_info_description_mismatched_type(self):
        path = self.get_contract_path('MetadataInfoDescriptionMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_metadata_info_extras(self):
        expected_output = (
            Opcode.PUSH5  # return 5
            + Opcode.RET
        )

        path = self.get_contract_path('MetadataInfoExtras.py')
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        self.assertIn('extra', manifest)
        self.assertIsNotNone(manifest['extra'])
        self.assertIn('Author', manifest['extra'])
        self.assertEqual('Test', manifest['extra']['Author'])
        self.assertIn('Email', manifest['extra'])
        self.assertEqual('test@test.com', manifest['extra']['Email'])
        self.assertIn('Description', manifest['extra'])
        self.assertEqual('This is an example', manifest['extra']['Description'])

    def test_metadata_info_new_extras(self):
        expected_output = (
            Opcode.PUSH5  # return 5
            + Opcode.RET
        )

        path = self.get_contract_path('MetadataInfoNewExtras.py')
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)
        self.assertIn('extra', manifest)
        self.assertIsNotNone(manifest['extra'])
        self.assertIn('UnitTest1', manifest['extra'])
        self.assertEqual('string', manifest['extra']['UnitTest1'])
        self.assertIn('UnitTest2', manifest['extra'])
        self.assertEqual(123, manifest['extra']['UnitTest2'])
        self.assertIn('UnitTest3', manifest['extra'])
        self.assertEqual(True, manifest['extra']['UnitTest3'])
        self.assertIn('UnitTest4', manifest['extra'])
        self.assertEqual(['list', 3210], manifest['extra']['UnitTest4'])

    def test_metadata_info_supported_standards(self):
        path = self.get_contract_path('MetadataInfoSupportedStandards.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('supportedstandards', manifest)
        self.assertIsInstance(manifest['supportedstandards'], list)
        self.assertGreater(len(manifest['supportedstandards']), 0)
        self.assertIn('NEP-17', manifest['supportedstandards'])

        engine = TestEngine()
        # verify using NeoManifestStruct
        nef, manifest = self.get_bytes_output(path)
        self.run_smart_contract(engine, path, 'Main')
        call_hash = engine.executed_script_hash.to_array()
        path = path.replace('.py', '.nef')

        get_contract_path = self.get_contract_path('test_sc/native_test/contractmanagement', 'GetContract.py')
        engine = TestEngine()
        engine.add_contract(path)

        result = self.run_smart_contract(engine, get_contract_path, 'main', call_hash)
        manifest_struct = NeoManifestStruct.from_json(manifest)
        self.assertEqual(manifest_struct, result[4])

    def test_metadata_info_supported_standards_with_imported_event(self):
        path = self.get_contract_path('MetadataInfoSupportedStandardsImportedEvent.py')
        output, manifest = self.get_output(path)

        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('events', abi)
        self.assertEqual(1, len(abi['events']))

        self.assertIn('name', abi['events'][0])
        self.assertEqual('Transfer', abi['events'][0]['name'])

    def test_metadata_info_supported_standards_filter(self):
        path = self.get_contract_path('MetadataInfoSupportedStandardsFilter.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('supportedstandards', manifest)
        self.assertIsInstance(manifest['supportedstandards'], list)
        self.assertGreater(len(manifest['supportedstandards']), 0)
        self.assertIn('NEP-17', manifest['supportedstandards'])

        # Using a dot is permitted and won't be removed by the filter
        self.assertIn('NEP-17.1', manifest['supportedstandards'])
        self.assertIn('NEP-17.1.2', manifest['supportedstandards'])
        self.assertIn('NEP-17.1.3.5.7.9.11.13', manifest['supportedstandards'])     # possible to use more than 1 dot

        # Every NEP needs to be in uppercase and be separated by a dash
        self.assertIn('NEP-100', manifest['supportedstandards'])
        self.assertIn('NEP-101', manifest['supportedstandards'])
        self.assertIn('NEP-102', manifest['supportedstandards'])

        # Standards that doesn't begin with NEP will not the filtered
        self.assertIn('not neo standard 1', manifest['supportedstandards'])

    def test_metadata_info_supported_standards_not_explicit(self):
        path = self.get_contract_path('MetadataInfoSupportedStandardsNotExplicit.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('supportedstandards', manifest)
        self.assertIsInstance(manifest['supportedstandards'], list)
        self.assertGreater(len(manifest['supportedstandards']), 0)
        self.assertIn('NEP-17', manifest['supportedstandards'])

    def test_metadata_info_supported_standards_not_explicit_incomplete_methods(self):
        path = self.get_contract_path('MetadataInfoSupportedStandardsNotExplicitIncompleteMethods.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('supportedstandards', manifest)
        self.assertIsInstance(manifest['supportedstandards'], list)
        self.assertEqual(len(manifest['supportedstandards']), 0)

    def test_metadata_info_supported_standards_not_explicit_incomplete_events(self):
        path = self.get_contract_path('MetadataInfoSupportedStandardsNotExplicitIncompleteEvents.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('supportedstandards', manifest)
        self.assertIsInstance(manifest['supportedstandards'], list)
        self.assertEqual(len(manifest['supportedstandards']), 0)

    def test_metadata_info_supported_standards_missing_implementations_nep17(self):
        path = self.get_contract_path('MetadataInfoSupportedStandardsMissingImplementationNEP17.py')
        self.assertCompilerLogs(CompilerError.MissingStandardDefinition, path)

    def test_metadata_info_supported_standards_missing_implementations_nep11(self):
        path = self.get_contract_path('MetadataInfoSupportedStandardsMissingImplementationNEP11.py')
        self.assertCompilerLogs(CompilerError.MissingStandardDefinition, path)

    def test_metadata_info_supported_standards_missing_implementations_nep11_optional_method(self):
        path = self.get_contract_path('MetadataInfoSupportedStandardsMissingImplementationNEP11OptionalMethods.py')
        self.assertCompilerLogs(CompilerError.MissingStandardDefinition, path)

    def test_metadata_info_supported_standards_mismatched_type(self):
        path = self.get_contract_path('MetadataInfoDescriptionMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_metadata_info_trusts(self):
        path = self.get_contract_path('MetadataInfoTrusts.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('trusts', manifest)
        self.assertIsInstance(manifest['trusts'], list)
        self.assertEqual(len(manifest['trusts']), 4)
        self.assertIn('0x1234567890123456789012345678901234567890', manifest['trusts'])
        self.assertIn('0x1234567890123456789012345678901234abcdef', manifest['trusts'])
        self.assertIn('035a928f201639204e06b4368b1a93365462a8ebbff0b8818151b74faab3a2b61a', manifest['trusts'])
        self.assertIn('03cdb067d930fd5adaa6c68545016044aaddec64ba39e548250eaea551172e535c', manifest['trusts'])

        engine = TestEngine()
        # verify using NeoManifestStruct
        nef, manifest = self.get_bytes_output(path)
        self.run_smart_contract(engine, path, 'Main')
        call_hash = engine.executed_script_hash.to_array()
        path = path.replace('.py', '.nef')

        get_contract_path = self.get_contract_path('test_sc/native_test/contractmanagement', 'GetContract.py')
        engine = TestEngine()
        engine.add_contract(path)

        result = self.run_smart_contract(engine, get_contract_path, 'main', call_hash)
        manifest_struct = NeoManifestStruct.from_json(manifest)

        result_trusts = result[4][6]

        # transform the str values to bytes values
        manifest_struct_trusts = []
        for item in manifest_struct[6]:
            if isinstance(item, str):
                if len(item) == 42:
                    from boa3.neo3.core.types import UInt160
                    item = UInt160.from_string(item).to_array()
                elif len(item) == 66:
                    item = bytes.fromhex(item)
            manifest_struct_trusts.append(item)

        # compare result from GetContract and NeoManifestStruct
        self.assertEqual(manifest_struct_trusts, result_trusts)

    def test_metadata_info_trusts_wildcard(self):
        path = self.get_contract_path('MetadataInfoTrustsWildcard.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('trusts', manifest)
        self.assertIsInstance(manifest['trusts'], list)
        self.assertEqual(len(manifest['trusts']), 1)
        self.assertIn(IMPORT_WILDCARD, manifest['trusts'])

        engine = TestEngine()
        # verify using NeoManifestStruct
        nef, manifest = self.get_bytes_output(path)
        self.run_smart_contract(engine, path, 'Main')
        call_hash = engine.executed_script_hash.to_array()
        path = path.replace('.py', '.nef')

        get_contract_path = self.get_contract_path('test_sc/native_test/contractmanagement', 'GetContract.py')
        engine = TestEngine()
        engine.add_contract(path)

        result = self.run_smart_contract(engine, get_contract_path, 'main', call_hash)
        manifest_struct = NeoManifestStruct.from_json(manifest)

        result_trusts = result[4][6]

        # TODO: change when TestEngine is updated
        self.assertEqual([''], result_trusts)

    def test_metadata_info_trusts_mismatched_types(self):
        path = self.get_contract_path('MetadataInfoTrustsMismatchedTypes.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('trusts', manifest)
        self.assertIsInstance(manifest['trusts'], list)
        self.assertEqual(len(manifest['trusts']), 0)

    def test_metadata_info_trusts_default(self):
        path = self.get_contract_path('MetadataInfoTrustsDefault.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('trusts', manifest)
        self.assertIsInstance(manifest['trusts'], list)
        self.assertEqual(len(manifest['trusts']), 0)

    def test_metadata_info_permissions(self):
        path = self.get_contract_path('MetadataInfoPermissions.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('permissions', manifest)
        self.assertIsInstance(manifest['permissions'], list)
        self.assertEqual(len(manifest['permissions']), 3)
        self.assertIn({"contract": "*", "methods": ['onNEP17Payment']}, manifest['permissions'])
        self.assertIn({"contract": "0x3846a4aa420d9831044396dd3a56011514cd10e3", "methods": ["get_object"]}, manifest['permissions'])
        self.assertIn({"contract": "0333b24ee50a488caa5deec7e021ff515f57b7993b93b45d7df901e23ee3004916", "methods": "*"}, manifest['permissions'])

        engine = TestEngine()
        # verify using NeoManifestStruct
        nef, manifest = self.get_bytes_output(path)
        self.run_smart_contract(engine, path, 'Main')
        call_hash = engine.executed_script_hash.to_array()
        path = path.replace('.py', '.nef')

        get_contract_path = self.get_contract_path('test_sc/native_test/contractmanagement', 'GetContract.py')
        engine = TestEngine()
        engine.add_contract(path)

        result = self.run_smart_contract(engine, get_contract_path, 'main', call_hash)
        manifest_struct = NeoManifestStruct.from_json(manifest)

        result_permissions = result[4][5]

        # casting the addresses to bytes values
        manifest_struct_permissions = []
        for item in manifest_struct[5]:
            contract = item[0]

            from boa3.neo3.core.types import UInt160
            if isinstance(contract, UInt160):
                contract = contract.to_array()
            manifest_struct_permissions.append([contract, item[1]])

        # compare result from GetContract and NeoManifestStruct
        self.assertEqual(manifest_struct_permissions, result_permissions)

    def test_metadata_info_permissions_mismatched_type(self):
        path = self.get_contract_path('MetadataInfoPermissionsMismatchedType.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('permissions', manifest)
        self.assertIsInstance(manifest['permissions'], list)
        self.assertEqual(len(manifest['permissions']), 1)
        self.assertIn({"contract": "*", "methods": "*"}, manifest['permissions'])

    def test_metadata_info_permissions_default(self):
        path = self.get_contract_path('MetadataInfoPermissionsDefault.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('permissions', manifest)
        self.assertIsInstance(manifest['permissions'], list)
        self.assertEqual(len(manifest['permissions']), 1)
        self.assertIn({"contract": "*", "methods": "*"}, manifest['permissions'])

    def test_metadata_info_permissions_wildcard(self):
        path = self.get_contract_path('MetadataInfoPermissionsWildcard.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('permissions', manifest)
        self.assertIsInstance(manifest['permissions'], list)
        self.assertEqual(len(manifest['permissions']), 1)
        self.assertIn({"contract": "*", "methods": "*"}, manifest['permissions'])

    def test_metadata_info_permissions_native_contract(self):
        path = self.get_contract_path('MetadataInfoPermissionsNativeContract.py')
        output, manifest = self.compile_and_save(path)

        expected_permission = {
            'contract': str(UInt160(constants.MANAGEMENT_SCRIPT)),
            'methods': ['update', 'destroy']
        }
        self.assertIn('permissions', manifest)
        self.assertIsInstance(manifest['permissions'], list)
        self.assertEqual(len(manifest['permissions']), 1)
        self.assertIn(expected_permission, manifest['permissions'])

    def test_metadata_info_name(self):
        path = self.get_contract_path('MetadataInfoName.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('name', manifest)
        self.assertIsInstance(manifest['name'], str)
        self.assertGreater(len(manifest['name']), 0)
        self.assertEqual((manifest['name']), "SmartContractCustomName")

    def test_metadata_info_name_default(self):
        path = self.get_contract_path('MetadataInfoNameDefault.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('name', manifest)
        self.assertIsInstance(manifest['name'], str)
        self.assertGreater(len(manifest['name']), 0)
        self.assertEqual((manifest['name']), "MetadataInfoNameDefault")

    def test_metadata_info_name_mismatched_type(self):
        path = self.get_contract_path('MetadataInfoNameMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_metadata_info_groups(self):
        path = self.get_contract_path('MetadataInfoGroups.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('groups', manifest)
        self.assertIsInstance(manifest['groups'], list)
        self.assertEqual(len(manifest['groups']), 1)
        self.assertIn({'pubkey': '031f64da8a38e6c1e5423a72ddd6d4fc4a777abe537e5cb5aa0425685cda8e063b',
                       'signature': '2p4uEy4pE3yj8jjmkhNrH3e0jI/w4WJCy3tTqlomSvCekM60tQ0zpmFfke+YOXa3tq/MlXLavqGUpNq/Pq3h5Q=='}, manifest['groups'])

        engine = TestEngine()
        # verify using NeoManifestStruct
        nef, manifest = self.get_bytes_output(path)
        self.run_smart_contract(engine, path, 'main')
        call_hash = engine.executed_script_hash.to_array()
        path = path.replace('.py', '.nef')

        get_contract_path = self.get_contract_path('test_sc/native_test/contractmanagement', 'GetContract.py')
        engine = TestEngine()
        engine.add_contract(path)

        result = self.run_smart_contract(engine, get_contract_path, 'main', call_hash)
        manifest_struct = NeoManifestStruct.from_json(manifest)

        result_groups = []
        for group in result[4][1]:
            dict_group = {'pubkey': group[0].hex()}
            import base64
            # getContract returns the signature without the base64 encoding
            dict_group['signature'] = base64.b64encode(group[1]).decode('utf-8')
            result_groups.append(dict_group)

        manifest_struct_groups = manifest_struct[1]

        self.assertEqual(manifest_struct_groups, result_groups)

    def test_metadata_info_groups_default(self):
        path = self.get_contract_path('MetadataInfoGroupsDefault.py')
        output, manifest = self.compile_and_save(path)

        self.assertIn('groups', manifest)
        self.assertIsInstance(manifest['groups'], list)
        self.assertEqual(len(manifest['groups']), 0)

    def test_metadata_info_source(self):
        path = self.get_contract_path('MetadataInfoSource.py')
        self.compile_and_save(path)

        nef_path = path.replace('.py', '.nef')
        with open(nef_path, mode='rb') as nef:
            from boa3.neo.contracts.neffile import NefFile
            generated_source = NefFile.deserialize(nef.read()).source

        self.assertEqual(generated_source, 'https://github.com/CityOfZion/neo3-boa')

    def test_metadata_info_source_default(self):
        path = self.get_contract_path('MetadataInfoSourceDefault.py')
        self.compile_and_save(path)

        nef_path = path.replace('.py', '.nef')
        with open(nef_path, mode='rb') as nef:
            from boa3.neo.contracts.neffile import NefFile
            generated_source = NefFile.deserialize(nef.read()).source

        self.assertEqual(generated_source, '')

    def test_metadata_info_source_mismatched_type(self):
        path = self.get_contract_path('MetadataInfoSourceMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)



__all__ = ['AddressVersionProperty',
           'BlockTimeProperty',
           'BurnGasMethod',
           'CallingScriptHashProperty',
           'CheckWitnessMethod',
           'EntryScriptHashProperty',
           'ExecutingScriptHashProperty',
           'GasLeftProperty',
           'GetNetworkMethod',
           'GetNotificationsMethod',
           'GetRandomMethod',
           'InvocationCounterProperty',
           'LoadScriptMethod',
           'LogMethod',
           'NotificationType',
           'NotifyMethod',
           'PlatformProperty',
           'ScriptContainerProperty',
           'GetTriggerMethod',
           'TriggerType',
           ]

from boa3.model.builtin.interop.runtime.burngasmethod import BurnGasMethod
from boa3.model.builtin.interop.runtime.checkwitnessmethod import CheckWitnessMethod
from boa3.model.builtin.interop.runtime.getaddressversionmethod import AddressVersionProperty
from boa3.model.builtin.interop.runtime.getblocktimemethod import BlockTimeProperty
from boa3.model.builtin.interop.runtime.getcallingscripthashmethod import CallingScriptHashProperty
from boa3.model.builtin.interop.runtime.getentryscripthashmethod import EntryScriptHashProperty
from boa3.model.builtin.interop.runtime.getexecutingscripthashmethod import ExecutingScriptHashProperty
from boa3.model.builtin.interop.runtime.getgasleftmethod import GasLeftProperty
from boa3.model.builtin.interop.runtime.getinvocationcountermethod import InvocationCounterProperty
from boa3.model.builtin.interop.runtime.getnetworkmethod import GetNetworkMethod
from boa3.model.builtin.interop.runtime.getnotificationsmethod import GetNotificationsMethod
from boa3.model.builtin.interop.runtime.getplatformmethod import PlatformProperty
from boa3.model.builtin.interop.runtime.getrandommethod import GetRandomMethod
from boa3.model.builtin.interop.runtime.gettriggermethod import GetTriggerMethod
from boa3.model.builtin.interop.runtime.loadscriptmethod import LoadScriptMethod
from boa3.model.builtin.interop.runtime.logmethod import LogMethod
from boa3.model.builtin.interop.runtime.notificationtype import NotificationType
from boa3.model.builtin.interop.runtime.notifymethod import NotifyMethod
from boa3.model.builtin.interop.runtime.scriptcontainermethod import ScriptContainerProperty
from boa3.model.builtin.interop.runtime.triggertype import TriggerType



from boa3.boa3 import Boa3
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestOptional(BoaTest):
    default_folder: str = 'test_sc/optional_test'

    def test_optional_return(self):
        path = self.get_contract_path('OptionalReturn.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 1)
        self.assertEqual('str', result)
        result = self.run_smart_contract(engine, path, 'main', 2)
        self.assertEqual(123, result)
        result = self.run_smart_contract(engine, path, 'main', 3)
        self.assertEqual(None, result)

        result = self.run_smart_contract(engine, path, 'union_test', 1)
        self.assertEqual('str', result)
        result = self.run_smart_contract(engine, path, 'union_test', 2)
        self.assertEqual(123, result)
        result = self.run_smart_contract(engine, path, 'union_test', 3)
        self.assertEqual(None, result)

    def test_optional_variable_reassign(self):
        expected_output = (
            Opcode.INITSLOT  # function signature
            + b'\x03'
            + b'\x00'
            + Opcode.PUSH2  # a = 2
            + Opcode.STLOC0
            + Opcode.PUSH2  # b = a
            + Opcode.STLOC1
            + Opcode.PUSHNULL  # c = None
            + Opcode.STLOC2
            + Opcode.LDLOC2  # b = c
            + Opcode.STLOC1
            + Opcode.RET  # return
        )

        path = self.get_contract_path('OptionalVariableReassign.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_optional_variable_argument(self):
        path = self.get_contract_path('OptionalVariableArgument.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 'unittest')
        self.assertEqual('string', result)
        result = self.run_smart_contract(engine, path, 'main', 123)
        self.assertEqual('int', result)
        result = self.run_smart_contract(engine, path, 'main', None)
        self.assertEqual('None', result)

        result = self.run_smart_contract(engine, path, 'union_test', 'unittest')
        self.assertEqual('string', result)
        result = self.run_smart_contract(engine, path, 'union_test', 123)
        self.assertEqual('int', result)
        result = self.run_smart_contract(engine, path, 'union_test', None)
        self.assertEqual('None', result)

    def test_optional_isinstance_validation(self):
        path = self.get_contract_path('OptionalIsInstanceValidation.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', 'unittest')
        self.assertEqual('unittest', result)
        result = self.run_smart_contract(engine, path, 'main', 123)
        self.assertEqual('int', result)
        result = self.run_smart_contract(engine, path, 'main', None)
        self.assertEqual('None', result)

        result = self.run_smart_contract(engine, path, 'union_test', 'unittest')
        self.assertEqual('unittest', result)
        result = self.run_smart_contract(engine, path, 'union_test', 123)
        self.assertEqual('int', result)
        result = self.run_smart_contract(engine, path, 'union_test', None)
        self.assertEqual('None', result)



__all__ = ['AppendMethod',
           'ClearMethod',
           'CopyDictMethod',
           'CopyListMethod',
           'CountSequenceGenericMethod',
           'CountSequencePrimitiveMethod',
           'CountStrMethod',
           'ExtendMethod',
           'IndexBytesStringMethod',
           'IndexSequenceMethod',
           'InsertMethod',
           'IsDigitMethod',
           'JoinMethod',
           'LowerMethod',
           'MapKeysMethod',
           'MapValuesMethod',
           'PopDictMethod',
           'PopDictDefaultMethod',
           'PopSequenceMethod',
           'RemoveMethod',
           'ReverseMethod',
           'StartsWithMethod',
           'StripMethod',
           'UpperMethod',
           'ToBoolMethod',
           'ToBytesMethod',
           'ToIntMethod',
           'ToStrMethod',
           'UpperMethod',
           ]

from boa3.model.builtin.classmethod.appendmethod import AppendMethod
from boa3.model.builtin.classmethod.clearmethod import ClearMethod
from boa3.model.builtin.classmethod.copydictmethod import CopyDictMethod
from boa3.model.builtin.classmethod.copylistmethod import CopyListMethod
from boa3.model.builtin.classmethod.countsequencegenericmethod import CountSequenceGenericMethod
from boa3.model.builtin.classmethod.countsequenceprimitivemethod import CountSequencePrimitiveMethod
from boa3.model.builtin.classmethod.countstrmethod import CountStrMethod
from boa3.model.builtin.classmethod.extendmethod import ExtendMethod
from boa3.model.builtin.classmethod.indexbytesstringmethod import IndexBytesStringMethod
from boa3.model.builtin.classmethod.indexsequencemethod import IndexSequenceMethod
from boa3.model.builtin.classmethod.insertmethod import InsertMethod
from boa3.model.builtin.classmethod.isdigitmethod import IsDigitMethod
from boa3.model.builtin.classmethod.joinmethod import JoinMethod
from boa3.model.builtin.classmethod.lowermethod import LowerMethod
from boa3.model.builtin.classmethod.mapkeysmethod import MapKeysMethod
from boa3.model.builtin.classmethod.mapvaluesmethod import MapValuesMethod
from boa3.model.builtin.classmethod.popdictdefaultmethod import PopDictDefaultMethod
from boa3.model.builtin.classmethod.popdictmethod import PopDictMethod
from boa3.model.builtin.classmethod.popsequencemethod import PopSequenceMethod
from boa3.model.builtin.classmethod.removemethod import RemoveMethod
from boa3.model.builtin.classmethod.reversemethod import ReverseMethod
from boa3.model.builtin.classmethod.startswithmethod import StartsWithMethod
from boa3.model.builtin.classmethod.stripmethod import StripMethod
from boa3.model.builtin.classmethod.toboolmethod import ToBool as ToBoolMethod
from boa3.model.builtin.classmethod.tobytesmethod import ToBytes as ToBytesMethod
from boa3.model.builtin.classmethod.tointmethod import ToInt as ToIntMethod
from boa3.model.builtin.classmethod.tostrmethod import ToStr as ToStrMethod
from boa3.model.builtin.classmethod.uppermethod import UpperMethod



from typing import List

from boa3_test.tests.test_classes.testcontract import TestContract


class ContractCollection:
    def __init__(self):
        self._internal_list: List[TestContract] = []

    def append(self, new_contract: TestContract):
        return self._internal_list.append(new_contract)

    def remove(self, contract: TestContract):
        return self._internal_list.remove(contract)

    def clear(self):
        return self._internal_list.clear()

    def __len__(self):
        return len(self._internal_list)

    def __contains__(self, item) -> bool:
        return any(contract.is_valid_identifier(item) for contract in self._internal_list)

    def __getitem__(self, item) -> TestContract:
        return next((contract for contract in self._internal_list if contract.is_valid_identifier(item)), None)

    def __str__(self):
        return str(self._internal_list)

    def __repr__(self):
        return self._internal_list.__repr__()



from typing import Any

from boa3.builtin.compile_time import public
from boa3.builtin.interop.contract import update_contract
from boa3.builtin.interop.runtime import notify


@public
def update(script: bytes, manifest: bytes, data: Any):
    update_contract(script, manifest, data)


@public
def new_method() -> int:
    return 42


@public
def _deploy(data: Any, update: bool):
    notify(update)
    notify(data)



from typing import Any, List, Union, Sequence

from boa3.builtin.interop.contract.callflagstype import CallFlags
from boa3.builtin.interop.runtime.notification import Notification
from boa3.builtin.interop.runtime.triggertype import TriggerType
from boa3.builtin.type import ECPoint, UInt160, ByteString


def check_witness(hash_or_pubkey: Union[UInt160, ECPoint]) -> bool:
    """
    Verifies that the transactions or block of the calling contract has validated the required script hash.

    :param hash_or_pubkey: script hash or public key to validate
    :type hash_or_pubkey: UInt160 or ECPoint
    :return: a boolean value that represents whether the script hash was verified
    :rtype: bool
    """
    pass


def notify(state: Any, notification_name: str = None):
    """
    Notifies the client from the executing smart contract.

    :param state: the notification message
    :type state: Any
    :param notification_name: name that'll be linked to the notification
    :type notification_name: str
    """
    pass


def log(message: str):
    """
    Show log messages to the client from the executing smart contract.

    :param message: the log message
    :type message: str
    """
    pass


def get_trigger() -> TriggerType:
    """
    Return the smart contract trigger type.

    :return: a value that represents the contract trigger type
    :rtype: TriggerType
    """
    pass


def get_notifications(script_hash: UInt160 = UInt160()) -> List[Notification]:
    """
    This method gets current invocation notifications from specific 'script_hash'.

    :param script_hash: must have 20 bytes, but if it's all zero 0000...0000 it refers to all existing notifications
        (like a * wildcard)
    :type script_hash: UInt160
    :return: It will return an array of all matched notifications
    :rtype: List[Notification]
    """
    pass


def get_network() -> int:
    """
    Gets the magic number of the current network.

    :return: the magic number of the current network
    :rtype: int
    """
    pass


def burn_gas(gas: int):
    """
    Burns GAS to benefit the NEO ecosystem.

    :param gas: the amount of GAS that will be burned
    :type gas: int

    :raise Exception: raised if gas value is negative.
    """
    pass


def get_random() -> int:
    """
    Gets the next random number.

    :return: the next random number
    :rtype: int
    """
    pass


def load_script(script: ByteString, args: Sequence = (), flags: CallFlags = CallFlags.NONE) -> Any:
    """
    Loads a script at runtime.
    """
    pass


address_version: int = 0
"""
Gets the address version of the current network.

:meta hide-value:
"""


executing_script_hash: UInt160 = UInt160()
"""
Gets the script hash of the current context.

:meta hide-value:
"""

calling_script_hash: UInt160 = UInt160()
"""
Gets the script hash of the calling contract.

:meta hide-value:
"""

time: int = 0
"""
Gets the timestamp of the current block.

:meta hide-value:
"""

gas_left: int = 0
"""
Gets the remaining GAS that can be spent in order to complete the execution.

:meta hide-value:
"""

platform: str = ''
"""
Gets the name of the current platform.

:meta hide-value:
"""

invocation_counter: int = 0
"""
Gets the number of times the current contract has been called during the execution.

:meta hide-value:
"""

entry_script_hash: UInt160 = UInt160()
"""
Gets the script hash of the entry context.

:meta hide-value:
"""

script_container: Any = None
"""
Gets the current script container.

:meta hide-value:
"""



import ast
from typing import Optional

from boa3.model import set_internal_call
from boa3.model.builtin.decorator.builtindecorator import IBuiltinDecorator
from boa3.model.symbol import ISymbol
from boa3.model.type.annotation import metatype
from boa3.model.type.classes.userclass import UserClass


class PropertyDecorator(IBuiltinDecorator):
    def __init__(self):
        identifier = 'property'
        super().__init__(identifier)

        """
        @property
        def property_name(self):
            return something
        """

    def update_args(self, args: ast.arguments, origin: Optional[ISymbol] = None):
        if isinstance(origin, UserClass) and len(args.args) == 1 and args.args[0].annotation is None:
            # the user doesn't need to explicitly write the type of the first argument if it's a classmethod
            # the first argument is a Type[Class]
            cls_type = metatype.metaType.build(origin)
            cls_type_annotation = (cls_type.meta_identifier
                                   if isinstance(cls_type, metatype.MetaType)
                                   else cls_type.identifier)

            cls_ast_annotation = ast.parse(cls_type_annotation).body[0].value
            cls_ast_annotation = set_internal_call(cls_ast_annotation)
            args.args[0].annotation = cls_ast_annotation

    @property
    def has_cls_or_self(self) -> bool:
        return True



from boa3.model.type.primitive.strtype import StrType
from boa3.neo.vm.type.StackItem import StackItemType


class BufferType(StrType):
    """
    A class used to represent Neo internal Buffer type
    """

    def __init__(self):
        super().__init__()
        self._identifier = 'buffer'

    @property
    def stack_item(self) -> StackItemType:
        return StackItemType.Buffer


Buffer = BufferType()



from typing import Dict

from boa3.model.builtin.builtinproperty import IBuiltinProperty
from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.variable import Variable


class GetExecutingScriptHashMethod(InteropMethod):
    def __init__(self):
        from boa3.model.type.collection.sequence.uint160type import UInt160Type
        identifier = '-get_executing_script_hash'
        syscall = 'System.Runtime.GetExecutingScriptHash'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, syscall, args, return_type=UInt160Type.build())


class ExecutingScriptHashProperty(IBuiltinProperty):
    def __init__(self):
        identifier = 'executing_script_hash'
        getter = GetExecutingScriptHashMethod()
        super().__init__(identifier, getter)



from boa3.exception import CompilerError
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestMath(BoaTest):

    default_folder: str = 'test_sc/math_test'

    def test_no_import(self):
        path = self.get_contract_path('NoImport.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    # region pow test

    def test_pow_method(self):
        path = self.get_contract_path('Pow.py')
        engine = TestEngine()

        import math

        base = 1
        exponent = 4
        result = self.run_smart_contract(engine, path, 'main', base, exponent)
        self.assertEqual(math.pow(base, exponent), result)

        base = 5
        exponent = 2
        result = self.run_smart_contract(engine, path, 'main', base, exponent)
        self.assertEqual(math.pow(base, exponent), result)

        base = -2
        exponent = 2
        result = self.run_smart_contract(engine, path, 'main', base, exponent)
        self.assertEqual(math.pow(base, exponent), result)

        base = -2
        exponent = 3
        result = self.run_smart_contract(engine, path, 'main', base, exponent)
        self.assertEqual(math.pow(base, exponent), result)

        base = 2
        exponent = 0
        result = self.run_smart_contract(engine, path, 'main', base, exponent)
        self.assertEqual(math.pow(base, exponent), result)

    def test_pow_method_from_math(self):
        path = self.get_contract_path('PowFromMath.py')
        engine = TestEngine()

        from math import pow

        base = 2
        exponent = 3
        result = self.run_smart_contract(engine, path, 'main', base, exponent)
        self.assertEqual(pow(base, exponent), result)

    # endregion

    # region sqrt test

    def test_sqrt_method(self):
        path = self.get_contract_path('Sqrt.py')
        engine = TestEngine()

        from math import sqrt

        expected_result = int(sqrt(0))
        result = self.run_smart_contract(engine, path, 'main', 0)
        self.assertEqual(expected_result, result)
        expected_result = int(sqrt(1))
        result = self.run_smart_contract(engine, path, 'main', 1)
        self.assertEqual(expected_result, result)
        expected_result = int(sqrt(3))
        result = self.run_smart_contract(engine, path, 'main', 3)
        self.assertEqual(expected_result, result)
        expected_result = int(sqrt(4))
        result = self.run_smart_contract(engine, path, 'main', 4)
        self.assertEqual(expected_result, result)
        expected_result = int(sqrt(8))
        result = self.run_smart_contract(engine, path, 'main', 8)
        self.assertEqual(expected_result, result)
        expected_result = int(sqrt(10))
        result = self.run_smart_contract(engine, path, 'main', 10)
        self.assertEqual(expected_result, result)

        with self.assertRaisesRegex(TestExecutionException, self.VALUE_CANNOT_BE_NEGATIVE_MSG):
            self.run_smart_contract(engine, path, 'main', -1)

        val = 25
        expected_result = int(sqrt(val))
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(expected_result, result)

    def test_sqrt_method_from_math(self):
        path = self.get_contract_path('SqrtFromMath.py')
        engine = TestEngine()

        from math import sqrt

        val = 25
        expected_result = int(sqrt(val))
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(expected_result, result)

    # endregion



from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.exception.NotLoadedException import NotLoadedException
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestClass(BoaTest):
    default_folder: str = 'test_sc/class_test'

    def test_notification_get_variables(self):
        path = self.get_contract_path('NotificationGetVariables.py')
        output, manifest = self.compile_and_save(path)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'script_hash', [],
                                         expected_result_type=bytes)
        script = engine.executed_script_hash.to_array()

        contract_notifications = engine.get_events(origin=engine.executed_script_hash)
        self.assertEqual(len(contract_notifications), 0)
        self.assertEqual(bytes(20), result)

        result = self.run_smart_contract(engine, path, 'event_name', [])
        contract_notifications = engine.get_events(origin=engine.executed_script_hash)
        self.assertEqual(len(contract_notifications), 0)
        self.assertEqual('', result)

        result = self.run_smart_contract(engine, path, 'state', [])
        contract_notifications = engine.get_events(origin=engine.executed_script_hash)
        self.assertEqual(len(contract_notifications), 0)
        self.assertEqual([], result)

        result = self.run_smart_contract(engine, path, 'script_hash', [1])
        contract_notifications = engine.get_events(origin=engine.executed_script_hash)
        self.assertEqual(len(contract_notifications), 1)
        self.assertEqual(script, result)

        engine.reset_engine()
        result = self.run_smart_contract(engine, path, 'event_name', [1])
        contract_notifications = engine.get_events(origin=engine.executed_script_hash)
        self.assertEqual(len(contract_notifications), 1)
        self.assertEqual('notify', result)

        engine.reset_engine()
        result = self.run_smart_contract(engine, path, 'state', [1])
        contract_notifications = engine.get_events(origin=engine.executed_script_hash)
        self.assertEqual(len(contract_notifications), 1)
        self.assertEqual([1], result)

        engine.reset_engine()
        result = self.run_smart_contract(engine, path, 'state', ['1'])
        contract_notifications = engine.get_events(origin=engine.executed_script_hash)
        self.assertEqual(len(contract_notifications), 1)
        self.assertEqual(['1'], result)

    def test_notification_set_variables(self):
        path = self.get_contract_path('NotificationSetVariables.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'script_hash', b'',
                                         expected_result_type=bytes)
        self.assertEqual(b'', result)
        script = engine.executed_script_hash.to_array()

        result = self.run_smart_contract(engine, path, 'script_hash', script,
                                         expected_result_type=bytes)
        self.assertEqual(script, result)

        result = self.run_smart_contract(engine, path, 'event_name', 'unit test')
        self.assertEqual('unit test', result)

        result = self.run_smart_contract(engine, path, 'state', (1, 2, 3))
        self.assertEqual([1, 2, 3], result)

    def test_contract_constructor(self):
        path = self.get_contract_path('ContractConstructor.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'new_contract')
        self.assertEqual(5, len(result))

        if isinstance(result[2], str):
            result[2] = String(result[2]).to_bytes()
        if isinstance(result[3], str):
            result[3] = String(result[3]).to_bytes()

        self.assertEqual(0, result[0])
        self.assertEqual(0, result[1])
        self.assertEqual(bytes(20), result[2])
        self.assertEqual(bytes(), result[3])
        self.assertEqual({}, result[4])

    def test_user_class_empty(self):
        # since 0.11.2 methods that are not public nor are called are not generated to optimize code
        # this test generates an empty contract
        path = self.get_contract_path('UserClassEmpty.py')

        with self.assertRaises(NotLoadedException) as e:
            output = Boa3.compile(path)

        self.assertTrue(e.exception.empty_script)

    def test_user_class_with_static_method_from_class(self):
        path = self.get_contract_path('UserClassWithStaticMethodFromClass.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'call_by_class_name')
        self.assertEqual(42, result)

    def test_user_class_with_static_method_from_class_with_same_method_name(self):
        path = self.get_contract_path('UserClassWithStaticMethodFromClassWithSameNameMethod.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'call_by_class_name')
        self.assertEqual(42, result)

    def test_user_class_with_static_method_from_object(self):
        path = self.get_contract_path('UserClassWithStaticMethodFromObject.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_user_class_with_static_method_with_args(self):
        path = self.get_contract_path('UserClassWithStaticMethodWithArgs.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'call_by_class_name', 10, 10)
        self.assertEqual(30, result)

        result = self.run_smart_contract(engine, path, 'call_by_class_name', 30, -30)
        self.assertEqual(10, result)

        result = self.run_smart_contract(engine, path, 'call_by_class_name', -5, -10)
        self.assertEqual(-5, result)

    def test_user_class_with_static_method_with_vararg(self):
        path = self.get_contract_path('UserClassWithStaticMethodWithVararg.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'call_by_class_name', [])
        self.assertEqual(42, result)

        args = [1, 2, 3]
        result = self.run_smart_contract(engine, path, 'call_by_class_name', args)
        self.assertEqual(args[0], result)

        args = [4, 3, 2, 1]
        result = self.run_smart_contract(engine, path, 'call_by_class_name', args)
        self.assertEqual(args[0], result)

    def test_user_class_with_static_method_not_class_method(self):
        path = self.get_contract_path('UserClassWithStaticMethodNotClassMethod.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'call_by_class_name', 42)
        self.assertEqual(42, result)

    def test_user_class_with_class_method_called_from_class_name(self):
        path = self.get_contract_path('UserClassWithClassMethodFromClass.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'call_by_class_name')
        self.assertEqual(42, result)

    def test_user_class_with_class_method_called_from_object(self):
        path = self.get_contract_path('UserClassWithClassMethodFromObject.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'call_by_class_name')
        self.assertEqual(42, result)

    def test_user_class_with_class_method_called_from_variable(self):
        path = self.get_contract_path('UserClassWithClassMethodFromVariable.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'call_by_class_name')
        self.assertEqual(42, result)

    def test_user_class_with_class_method_with_args(self):
        path = self.get_contract_path('UserClassWithClassMethodWithArgs.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'call_by_class_name', 42)
        self.assertEqual(42, result)

        result = self.run_smart_contract(engine, path, 'call_by_class_name', 1)
        self.assertEqual(1, result)

        result = self.run_smart_contract(engine, path, 'call_by_class_name', -10)
        self.assertEqual(-10, result)

    def test_user_class_with_class_method_with_vararg(self):
        path = self.get_contract_path('UserClassWithClassMethodWithVararg.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'call_by_class_name', [])
        self.assertEqual(42, result)

        args = [1, 2, 3]
        result = self.run_smart_contract(engine, path, 'call_by_class_name', args)
        self.assertEqual(args[0], result)

        args = [4, 3, 2, 1]
        result = self.run_smart_contract(engine, path, 'call_by_class_name', args)
        self.assertEqual(args[0], result)

    def test_user_class_with_class_variable_from_class(self):
        path = self.get_contract_path('UserClassWithClassVariableFromClass.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'get_val1')
        self.assertEqual(1, result)

        result = self.run_smart_contract(engine, path, 'get_val2')
        self.assertEqual(2, result)

    def test_user_class_with_class_variable_from_object(self):
        path = self.get_contract_path('UserClassWithClassVariableFromObject.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'get_val1')
        self.assertEqual(1, result)

        result = self.run_smart_contract(engine, path, 'get_val2')
        self.assertEqual(2, result)

    def test_user_class_with_class_variable_from_variable(self):
        path = self.get_contract_path('UserClassWithClassVariableFromVariable.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'get_val1')
        self.assertEqual(1, result)

        result = self.run_smart_contract(engine, path, 'get_val2')
        self.assertEqual(2, result)

    def test_user_class_update_class_variable(self):
        path = self.get_contract_path('UserClassUpdateClassVariable.py')
        self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)

    def test_user_class_update_instance_variable_on_init(self):
        path = self.get_contract_path('UserClassUpdateClassVariableOnInit.py')
        self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)

    def test_user_class_with_class_variable_and_class_method(self):
        path = self.get_contract_path('UserClassWithClassVariableAndClassMethod.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'get_val1')
        self.assertEqual(1, result)

        result = self.run_smart_contract(engine, path, 'get_foo')
        self.assertEqual(42, result)

    def test_user_class_with_init(self):
        path = self.get_contract_path('UserClassWithInit.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'build_example_object')
        self.assertEqual([], result)

    def test_user_class_with_init_with_args(self):
        path = self.get_contract_path('UserClassWithInitWithArgs.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'build_example_object')
        self.assertEqual([], result)

    def test_user_class_with_instance_method(self):
        path = self.get_contract_path('UserClassWithInstanceMethod.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'call_by_class_name')
        self.assertEqual(42, result)

    def test_user_class_with_instance_method_from_variable(self):
        path = self.get_contract_path('UserClassWithInstanceMethodFromVariable.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'call_by_class_name')
        self.assertEqual(42, result)

    def test_user_class_with_instance_variable_from_class(self):
        path = self.get_contract_path('UserClassWithInstanceVariableFromClass.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_user_class_with_instance_variable_from_object(self):
        path = self.get_contract_path('UserClassWithInstanceVariableFromObject.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'get_val1')
        self.assertEqual(1, result)

        result = self.run_smart_contract(engine, path, 'get_val2')
        self.assertEqual(2, result)

    def test_user_class_with_instance_variable_from_variable(self):
        path = self.get_contract_path('UserClassWithInstanceVariableFromVariable.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'get_val1')
        self.assertEqual(1, result)

        result = self.run_smart_contract(engine, path, 'get_val2')
        self.assertEqual(2, result)

    def test_user_class_update_instance_variable(self):
        path = self.get_contract_path('UserClassUpdateInstanceVariable.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'get_val', 10)
        self.assertEqual([10, 10, 2], result)

        result = self.run_smart_contract(engine, path, 'get_val', 40)
        self.assertEqual([10, 40, 2], result)

    def test_user_class_access_variable_on_init(self):
        path = self.get_contract_path('UserClassAccessInstanceVariableOnInit.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'get_obj')
        self.assertEqual([2, 4], result)

    def test_user_class_access_variable_on_method(self):
        path = self.get_contract_path('UserClassAccessInstanceVariableOnMethod.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'get_val1')
        self.assertEqual(1, result)

        result = self.run_smart_contract(engine, path, 'get_val2')
        self.assertEqual(4, result)

    def test_user_class_with_base(self):
        path = self.get_contract_path('UserClassWithBuiltinBase.py')
        self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)

    def test_user_class_with_created_base(self):
        path = self.get_contract_path('UserClassWithCreatedBase.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'implemented_method')
        self.assertEqual(42, result)

        result = self.run_smart_contract(engine, path, 'inherited_method')
        self.assertEqual(42, result)

    def test_user_class_with_cascated_created_base(self):
        path = self.get_contract_path('UserClassWithCascadeCreatedBase.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'implemented_method')
        self.assertEqual(42, result)

        result = self.run_smart_contract(engine, path, 'inherited_method')
        self.assertEqual(42, result)

    def test_user_class_with_created_base_with_variable(self):
        path = self.get_contract_path('UserClassWithCreatedBaseWithVariables.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'implemented_variable')
        self.assertEqual(42, result)

        result = self.run_smart_contract(engine, path, 'inherited_variable')
        self.assertEqual(42, result)

        new_value = 10
        result = self.run_smart_contract(engine, path, 'update_variable', new_value)
        self.assertEqual(new_value, result)

        new_value = -10
        result = self.run_smart_contract(engine, path, 'update_variable', new_value)
        self.assertEqual(new_value, result)

        new_value = 10_000_000
        result = self.run_smart_contract(engine, path, 'update_variable', new_value)
        self.assertEqual(new_value, result)

    def test_user_class_with_created_base_with_args(self):
        path = self.get_contract_path('UserClassWithCreatedBaseWithArgs.py')
        engine = TestEngine()

        init_value = 42
        result = self.run_smart_contract(engine, path, 'implemented_var', init_value)
        self.assertEqual(init_value, result)

        result = self.run_smart_contract(engine, path, 'inherited_var', init_value)
        self.assertEqual(init_value, result)

        init_value = -42
        result = self.run_smart_contract(engine, path, 'implemented_var', init_value)
        self.assertEqual(init_value, result)

        result = self.run_smart_contract(engine, path, 'inherited_var', init_value)
        self.assertEqual(init_value, result)

        init_value = 10_000_000
        result = self.run_smart_contract(engine, path, 'implemented_var', init_value)
        self.assertEqual(init_value, result)

        result = self.run_smart_contract(engine, path, 'inherited_var', init_value)
        self.assertEqual(init_value, result)

    def test_user_class_with_created_base_with_init(self):
        path = self.get_contract_path('UserClassWithCreatedBaseWithInit.py')
        engine = TestEngine()

        expected_result = 42
        result = self.run_smart_contract(engine, path, 'inherited_var')
        self.assertEqual(expected_result, result)

    def test_user_class_with_created_base_with_init_with_args(self):
        path = self.get_contract_path('UserClassWithCreatedBaseWithInitWithArgs.py')
        engine = TestEngine()

        expected_result = -10
        result = self.run_smart_contract(engine, path, 'inherited_var')
        self.assertEqual(expected_result, result)

    def test_user_class_with_created_base_with_more_variables(self):
        path = self.get_contract_path('UserClassWithCreatedBaseWithMoreVariables.py')
        engine = TestEngine()

        expected_result = [42, 10, 20]
        result = self.run_smart_contract(engine, path, 'get_full_object')
        self.assertEqual(expected_result, result)

    def test_user_class_with_created_base_with_more_variables_without_super_init(self):
        path = self.get_contract_path('UserClassWithCreatedBaseWithMoreVariablesWithoutSuperInit.py')
        self.assertCompilerLogs(CompilerError.MissingInitCall, path)

    def test_user_class_with_multiple_bases(self):
        path = self.get_contract_path('UserClassWithMultipleBases.py')
        self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)

    def test_user_class_with_keyword_base(self):
        path = self.get_contract_path('UserClassWithKeywordBase.py')
        self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)

    def test_user_class_with_decorator(self):
        path = self.get_contract_path('UserClassWithDecorator.py')
        self.assertCompilerLogs(CompilerError.NotSupportedOperation, path)

    def test_user_class_with_property_from_object(self):
        path = self.get_contract_path('UserClassWithPropertyFromObject.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'get_property')
        self.assertEqual(1, result)

    def test_user_class_with_property_using_instance_variables_from_object(self):
        path = self.get_contract_path('UserClassWithPropertyUsingInstanceVariablesFromObject.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'get_property')
        self.assertEqual(10, result)

    def test_user_class_with_property_using_class_variables_from_object(self):
        path = self.get_contract_path('UserClassWithPropertyUsingClassVariablesFromObject.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'get_property')
        self.assertEqual(10, result)

    def test_user_class_with_property_using_variables_from_object(self):
        path = self.get_contract_path('UserClassWithPropertyUsingVariablesFromObject.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'get_property')
        self.assertEqual(47, result)

    def test_user_class_with_property_from_class(self):
        path = self.get_contract_path('UserClassWithPropertyFromClass.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_user_class_with_property_using_arguments(self):
        path = self.get_contract_path('UserClassWithPropertyUsingArguments.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    def test_user_class_with_property_mismatched_type(self):
        path = self.get_contract_path('UserClassWithPropertyMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_user_class_with_property_without_self(self):
        path = self.get_contract_path('UserClassWithPropertyWithoutSelf.py')
        self.assertCompilerLogs(CompilerError.SelfArgumentError, path)

    def test_user_class_with_augmented_assignment_operator_with_variable(self):
        path = self.get_contract_path('UserClassWithAugmentedAssignmentOperatorWithVariable.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'add')
        self.assertEqual(2, result)

        result = self.run_smart_contract(engine, path, 'sub')
        self.assertEqual(-2, result)

        result = self.run_smart_contract(engine, path, 'mult')
        self.assertEqual(16, result)

        result = self.run_smart_contract(engine, path, 'div')
        self.assertEqual(2, result)

        result = self.run_smart_contract(engine, path, 'mod')
        self.assertEqual(1, result)

        result = self.run_smart_contract(engine, path, 'mix')
        self.assertEqual(0, result)

    def test_user_class_with_deploy_method(self):
        path = self.get_contract_path('UserClassWithDeployMethod.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'get_obj')
        self.assertEqual([1, 2], result)



import enum

from boa3.model.callable import Callable
from boa3.model.event import Event


class ManifestSymbol(int, enum.Enum):
    Method = enum.auto()
    Event = enum.auto()

    @classmethod
    def get_manifest_symbol(cls, callable_: Callable):
        if isinstance(callable_, Event):
            return cls.Event
        return cls.Method

    def __str__(self):
        return self.name

    def __repr__(self):
        return str(self)



from typing import Any, List, Optional, Tuple

from boa3 import constants
from boa3.model.method import Method
from boa3.model.type.itype import IType
from boa3.model.type.primitive.bytestype import BytesType
from boa3.neo.vm.opcode import OpcodeHelper
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.AbiType import AbiType


class UInt160Type(BytesType):
    """
    A class used to represent Neo's UInt160 type
    """

    def __init__(self):
        super().__init__()
        self._identifier = 'UInt160'
        from boa3.model.builtin.method.uint160method import UInt160Method
        self._constructor = UInt160Method(self)

    @property
    def identifier(self) -> str:
        return self._identifier

    @property
    def abi_type(self) -> AbiType:
        return AbiType.Hash160

    def constructor_method(self) -> Optional[Method]:
        return self._constructor

    @property
    def default_value(self) -> Any:
        return bytes(constants.SIZE_OF_INT160)

    @classmethod
    def build(cls, value: Any = None) -> IType:
        return _UInt160

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, UInt160Type)

    def is_instance_opcodes(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.model.type.classes.pythonclass import PythonClass
        return super(PythonClass, self).is_instance_opcodes()

    def _is_instance_inner_opcodes(self, jmp_to_if_false: int = 0) -> List[Tuple[Opcode, bytes]]:
        push_int_opcode, size_data = OpcodeHelper.get_push_and_data(constants.SIZE_OF_INT160)

        return [
            (Opcode.SIZE, b''),  # return len(value) == 20
            (push_int_opcode, size_data),
            (Opcode.NUMEQUAL, b'')
        ]


_UInt160 = UInt160Type()



from typing import Union, Any

from boa3.builtin.type import UInt160


class Oracle:
    """
    Neo Oracle Service is an out-of-chain data access service built into Neo N3. It allows users to request the external
    data sources in smart contracts, and Oracle nodes designated by the committee will access the specified data source
    then pass the result in the callback function to continue executing the smart contract logic.
    """
    hash: UInt160

    @classmethod
    def request(cls, url: str, request_filter: Union[str, None], callback: str, user_data: Any, gas_for_response: int):
        """
        Requests an information from outside the blockchain.

        This method just requests data from the oracle, it won't return the result.

        :param url: External url to retrieve the data
        :type url: str
        :param request_filter:
            Filter to the request.

            See JSONPath format https://github.com/atifaziz/JSONPath

        :type request_filter: str or None
        :param callback:
            Method name that will be as a callback.

            This method must be public and implement the following interface:

            ``(url: str, user_data: Any, code: int, result: bytes) -> None``

        :type callback: str
        :param user_data: Optional data. It'll be returned as the same when the callback is called
        :type user_data: Any
        :param gas_for_response:
            Amount of GAS needed to run the callback method.

            It MUST NOT be specified as the user representation.

            If it costs 1 gas, this value must be 1_00000000 (with the 8 decimals)
        :type gas_for_response: int
        """
        pass

    @classmethod
    def get_price(cls) -> int:
        """
        Gets the price for an Oracle request.

        :return: the price for an Oracle request
        """
        pass



from __future__ import annotations

from enum import Enum


class Opcode(bytes, Enum):
    """
    """

    # region Constants

    PUSHINT8 = b'\x00'
    """
    Pushes a 1-byte signed integer onto the stack.

    :meta hide-value:
    """
    PUSHINT16 = b'\x01'
    """
    Pushes a 2-byte signed integer onto the stack.

    :meta hide-value:
    """
    PUSHINT32 = b'\x02'
    """
    Pushes a 4-byte signed integer onto the stack.

    :meta hide-value:
    """
    PUSHINT64 = b'\x03'
    """
    Pushes a 8-byte signed integer onto the stack.

    :meta hide-value:
    """
    PUSHINT128 = b'\x04'
    """
    Pushes a 16-byte signed integer onto the stack.

    :meta hide-value:
    """
    PUSHINT256 = b'\x05'
    """
    Pushes a 32-byte signed integer onto the stack.

    :meta hide-value:
    """

    PUSHT = b'\x08'
    """
    Pushes the boolean value True onto the stack.

    :meta hide-value:
    """
    PUSHF = b'\x09'
    """
    Pushes the boolean value False onto the stack.

    :meta hide-value:
    """
    PUSHA = b'\x0A'
    """
    Converts the 4-bytes offset to a Pointer, and pushes it onto the stack.

    :meta hide-value:
    """
    PUSHNULL = b'\x0B'
    """
    The item null is pushed onto the stack.

    :meta hide-value:
    """
    PUSHDATA1 = b'\x0C'
    """
    The next byte contains the number of bytes to be pushed onto the stack.

    :meta hide-value:
    """
    PUSHDATA2 = b'\x0D'
    """
    The next two bytes contain the number of bytes to be pushed onto the stack.

    :meta hide-value:
    """
    PUSHDATA4 = b'\x0E'
    """
    The next four bytes contain the number of bytes to be pushed onto the stack.

    :meta hide-value:
    """

    PUSHM1 = b'\x0F'
    """
    The number -1 is pushed onto the stack.

    :meta hide-value:
    """
    PUSH0 = b'\x10'
    """
    The number 0 is pushed onto the stack.

    :meta hide-value:
    """
    PUSH1 = b'\x11'
    """
    The number 1 is pushed onto the stack.

    :meta hide-value:
    """
    PUSH2 = b'\x12'
    """
    The number 2 is pushed onto the stack.

    :meta hide-value:
    """
    PUSH3 = b'\x13'
    """
    The number 3 is pushed onto the stack.

    :meta hide-value:
    """
    PUSH4 = b'\x14'
    """
    The number 4 is pushed onto the stack.

    :meta hide-value:
    """
    PUSH5 = b'\x15'
    """
    The number 5 is pushed onto the stack.

    :meta hide-value:
    """
    PUSH6 = b'\x16'
    """
    The number 6 is pushed onto the stack.

    :meta hide-value:
    """
    PUSH7 = b'\x17'
    """
    The number 7 is pushed onto the stack.

    :meta hide-value:
    """
    PUSH8 = b'\x18'
    """
    The number 8 is pushed onto the stack.

    :meta hide-value:
    """
    PUSH9 = b'\x19'
    """
    The number 9 is pushed onto the stack.

    :meta hide-value:
    """
    PUSH10 = b'\x1A'
    """
    The number 10 is pushed onto the stack.

    :meta hide-value:
    """
    PUSH11 = b'\x1B'
    """
    The number 11 is pushed onto the stack.

    :meta hide-value:
    """
    PUSH12 = b'\x1C'
    """
    The number 12 is pushed onto the stack.

    :meta hide-value:
    """
    PUSH13 = b'\x1D'
    """
    The number 13 is pushed onto the stack.

    :meta hide-value:
    """
    PUSH14 = b'\x1E'
    """
    The number 14 is pushed onto the stack.

    :meta hide-value:
    """
    PUSH15 = b'\x1F'
    """
    The number 15 is pushed onto the stack.

    :meta hide-value:
    """
    PUSH16 = b'\x20'
    """
    The number 16 is pushed onto the stack.

    :meta hide-value:
    """

    # endregion

    # region Flow control

    NOP = b'\x21'
    """
    The NOP operation does nothing. It is intended to fill in space if opcodes are patched.

    :meta hide-value:
    """
    JMP = b'\x22'
    """
    Unconditionally transfers control to a target instruction. The target instruction is represented as a 1-byte
    signed offset from the beginning of the current instruction.

    :meta hide-value:
    """
    JMP_L = b'\x23'
    """
    Unconditionally transfers control to a target instruction. The target instruction is represented as a 4-bytes
    signed offset from the beginning of the current instruction.

    :meta hide-value:
    """
    JMPIF = b'\x24'
    """
    Transfers control to a target instruction if the value is True, not null, or non-zero. The target instruction
    is represented as a 1-byte signed offset from the beginning of the current instruction.

    :meta hide-value:
    """
    JMPIF_L = b'\x25'
    """
    Transfers control to a target instruction if the value is True, not null, or non-zero. The target instruction
    is represented as a 4-bytes signed offset from the beginning of the current instruction.

    :meta hide-value:
    """
    JMPIFNOT = b'\x26'
    """
    Transfers control to a target instruction if the value is False, a null reference,
    or zero. The target instruction is represented as a 1-byte signed offset from the beginning of the current
    instruction.

    :meta hide-value:
    """
    JMPIFNOT_L = b'\x27'
    """
    Transfers control to a target instruction if the value is False, a null reference,
    or zero. The target instruction is represented as a 4-bytes signed offset from the beginning of the current
    instruction.

    :meta hide-value:
    """
    JMPEQ = b'\x28'
    """
    Transfers control to a target instruction if two values are equal. The target instruction is represented as a
    1-byte signed offset from the beginning of the current instruction.

    :meta hide-value:
    """
    JMPEQ_L = b'\x29'
    """
    Transfers control to a target instruction if two values are equal. The target instruction is represented as a
    4-bytes signed offset from the beginning of the current instruction.

    :meta hide-value:
    """
    JMPNE = b'\x2A'
    """
    Transfers control to a target instruction when two values are not equal. The target instruction is represented
    as a 1-byte signed offset from the beginning of the current instruction.

    :meta hide-value:
    """
    JMPNE_L = b'\x2B'
    """
    Transfers control to a target instruction when two values are not equal. The target instruction is represented
    as a 4-bytes signed offset from the beginning of the current instruction.

    :meta hide-value:
    """
    JMPGT = b'\x2C'
    """
    Transfers control to a target instruction if the first value is greater than the second value. The target
    instruction is represented as a 1-byte signed offset from the beginning of the current instruction.

    :meta hide-value:
    """
    JMPGT_L = b'\x2D'
    """
    Transfers control to a target instruction if the first value is greater than the second value. The target
    instruction is represented as a 4-bytes signed offset from the beginning of the current instruction.

    :meta hide-value:
    """
    JMPGE = b'\x2E'
    """
    Transfers control to a target instruction if the first value is greater than or equal to the second value. The
    target instruction is represented as a 1-byte signed offset from the beginning of the current instruction.

    :meta hide-value:
    """
    JMPGE_L = b'\x2F'
    """
    Transfers control to a target instruction if the first value is greater than or equal to the second value. The
    target instruction is represented as a 4-bytes signed offset from the beginning of the current instruction.

    :meta hide-value:
    """
    JMPLT = b'\x30'
    """
    Transfers control to a target instruction if the first value is less than the second value. The target
    instruction is represented as a 1-byte signed offset from the beginning of the current instruction.

    :meta hide-value:
    """
    JMPLT_L = b'\x31'
    """
    Transfers control to a target instruction if the first value is less than the second value. The target
    instruction is represented as a 4-bytes signed offset from the beginning of the current instruction.

    :meta hide-value:
    """
    JMPLE = b'\x32'
    """
    Transfers control to a target instruction if the first value is less than or equal to the second value. The
    target instruction is represented as a 1-byte signed offset from the beginning of the current instruction.

    :meta hide-value:
    """
    JMPLE_L = b'\x33'
    """
    Transfers control to a target instruction if the first value is less than or equal to the second value. The
    target instruction is represented as a 4-bytes signed offset from the beginning of the current instruction.

    :meta hide-value:
    """

    CALL = b'\x34'
    """
    Calls the function at the target address which is represented as a 1-byte signed offset from the beginning of
    the current instruction.

    :meta hide-value:
    """
    CALL_L = b'\x35'
    """
    Calls the function at the target address which is represented as a 4-bytes signed offset from the beginning of
    the current instruction.

    :meta hide-value:
    """
    CALLA = b'\x36'
    """
    Pop the address of a function from the stack, and call the function.

    :meta hide-value:
    """
    CALLT = b'\x37'
    """
    Calls the function which is described by the token.

    :meta hide-value:
    """
    ABORT = b'\x38'
    """
    It turns the vm state to FAULT immediately, and cannot be caught.

    :meta hide-value:
    """
    ASSERT = b'\x39'
    """
    Pop the top value of the stack, if it false, then exit vm execution and set vm state to FAULT.

    :meta hide-value:
    """
    THROW = b'\x3A'
    """
    Pop the top value of the stack, and throw it.

    :meta hide-value:
    """
    TRY = b'\x3B'
    """
    TRY CatchOffset(sbyte) FinallyOffset(sbyte). If there's no catch body, set CatchOffset 0. If there's no finally
    body, set FinallyOffset 0.

    :meta hide-value:
    """
    TRY_L = b'\x3C'
    """
    TRY_L CatchOffset(int) FinallyOffset(int). If there's no catch body, set CatchOffset 0. If there's no finally
    body, set FinallyOffset 0.

    :meta hide-value:
    """
    ENDTRY = b'\x3D'
    """
    Ensures that the appropriate surrounding finally blocks are executed. And then unconditionally transfers
    control to the specific target instruction, represented as a 1-byte signed offset from the beginning of the
    current instruction.

    :meta hide-value:
    """
    ENDTRY_L = b'\x3E'
    """
    Ensures that the appropriate surrounding finally blocks are executed. And then unconditionally transfers
    control to the specific target instruction, represented as a 4-byte signed offset from the beginning of the
    current instruction.

    :meta hide-value:
    """
    ENDFINALLY = b'\x3F'
    """
    End finally, If no exception happen or be catched, vm will jump to the target instruction of ENDTRY/ENDTRY_L.
    Otherwise vm will rethrow the exception to upper layer.

    :meta hide-value:
    """
    RET = b'\x40'
    """
    Returns from the current method.

    :meta hide-value:
    """
    SYSCALL = b'\x41'
    """
    Calls to an interop service.

    :meta hide-value:
    """

    # endregion

    # region Stack

    DEPTH = b'\x43'
    """
    Puts the number of stack items onto the stack.

    :meta hide-value:
    """
    DROP = b'\x45'
    """
    Removes the top stack item.

    :meta hide-value:
    """
    NIP = b'\x46'
    """
    Removes the second-to-top stack item.

    :meta hide-value:
    """
    XDROP = b'\x48'
    """
    The item n back in the main stack is removed.

    :meta hide-value:
    """
    CLEAR = b'\x49'
    """
    Clear the stack

    :meta hide-value:
    """

    DUP = b'\x4A'
    """
    Duplicates the top stack item.

    :meta hide-value:
    """
    OVER = b'\x4B'
    """
    Copies the second-to-top stack item to the top.

    :meta hide-value:
    """
    PICK = b'\x4D'
    """
    The item n back in the stack is copied to the top.

    :meta hide-value:
    """

    TUCK = b'\x4E'
    """
    The item at the top of the stack is copied and inserted before the second-to-top item.

    :meta hide-value:
    """
    SWAP = b'\x50'
    """
    The top two items on the stack are swapped.

    :meta hide-value:
    """
    ROT = b'\x51'
    """
    The top three items on the stack are rotated to the left.

    :meta hide-value:
    """
    ROLL = b'\x52'
    """
    The item n back in the stack is moved to the top.

    :meta hide-value:
    """
    REVERSE3 = b'\x53'
    """
    Reverse the order of the top 3 items on the stack.

    :meta hide-value:
    """
    REVERSE4 = b'\x54'
    """
    Reverse the order of the top 4 items on the stack.

    :meta hide-value:
    """
    REVERSEN = b'\x55'
    """
    Pop the number N on the stack, and reverse the order of the top N items on the stack.

    :meta hide-value:
    """

    # endregion

    # region Slot

    INITSSLOT = b'\x56'
    """
    Initialize the static field list for the current execution context.

    :meta hide-value:
    """
    INITSLOT = b'\x57'
    """
    Initialize the argument slot and the local variable list for the current execution context.

    :meta hide-value:
    """

    LDSFLD0 = b'\x58'
    """
    Loads the static field at index 0 onto the evaluation stack.

    :meta hide-value:
    """
    LDSFLD1 = b'\x59'
    """
    Loads the static field at index 1 onto the evaluation stack.

    :meta hide-value:
    """
    LDSFLD2 = b'\x5A'
    """
    Loads the static field at index 2 onto the evaluation stack.

    :meta hide-value:
    """
    LDSFLD3 = b'\x5B'
    """
    Loads the static field at index 3 onto the evaluation stack.

    :meta hide-value:
    """
    LDSFLD4 = b'\x5C'
    """
    Loads the static field at index 4 onto the evaluation stack.

    :meta hide-value:
    """
    LDSFLD5 = b'\x5D'
    """
    Loads the static field at index 5 onto the evaluation stack.

    :meta hide-value:
    """
    LDSFLD6 = b'\x5E'
    """
    Loads the static field at index 6 onto the evaluation stack.

    :meta hide-value:
    """
    LDSFLD = b'\x5F'
    """
    Loads the static field at a specified index onto the evaluation stack. The index is represented as a 1-byte
    unsigned integer.

    :meta hide-value:
    """
    STSFLD0 = b'\x60'
    """
    Stores the value on top of the evaluation stack in the static field list at index 0.

    :meta hide-value:
    """
    STSFLD1 = b'\x61'
    """
    Stores the value on top of the evaluation stack in the static field list at index 1.

    :meta hide-value:
    """
    STSFLD2 = b'\x62'
    """
    Stores the value on top of the evaluation stack in the static field list at index 2.

    :meta hide-value:
    """
    STSFLD3 = b'\x63'
    """
    Stores the value on top of the evaluation stack in the static field list at index 3.

    :meta hide-value:
    """
    STSFLD4 = b'\x64'
    """
    Stores the value on top of the evaluation stack in the static field list at index 4.

    :meta hide-value:
    """
    STSFLD5 = b'\x65'
    """
    Stores the value on top of the evaluation stack in the static field list at index 5.

    :meta hide-value:
    """
    STSFLD6 = b'\x66'
    """
    Stores the value on top of the evaluation stack in the static field list at index 6.

    :meta hide-value:
    """
    STSFLD = b'\x67'
    """
    Stores the value on top of the evaluation stack in the static field list at a specified index. The index is
    represented as a 1-byte unsigned integer.

    :meta hide-value:
    """
    LDLOC0 = b'\x68'
    """
    Loads the local variable at index 0 onto the evaluation stack.

    :meta hide-value:
    """
    LDLOC1 = b'\x69'
    """
    Loads the local variable at index 1 onto the evaluation stack.

    :meta hide-value:
    """
    LDLOC2 = b'\x6A'
    """
    Loads the local variable at index 2 onto the evaluation stack.

    :meta hide-value:
    """
    LDLOC3 = b'\x6B'
    """
    Loads the local variable at index 3 onto the evaluation stack.

    :meta hide-value:
    """
    LDLOC4 = b'\x6C'
    """
    Loads the local variable at index 4 onto the evaluation stack.

    :meta hide-value:
    """
    LDLOC5 = b'\x6D'
    """
    Loads the local variable at index 5 onto the evaluation stack.

    :meta hide-value:
    """
    LDLOC6 = b'\x6E'
    """
    Loads the local variable at index 6 onto the evaluation stack.

    :meta hide-value:
    """
    LDLOC = b'\x6F'
    """
    Loads the local variable at a specified index onto the evaluation stack. The index is represented as a 1-byte
    unsigned integer.

    :meta hide-value:
    """
    STLOC0 = b'\x70'
    """
    Stores the value on top of the evaluation stack in the local variable list at index 0.

    :meta hide-value:
    """
    STLOC1 = b'\x71'
    """
    Stores the value on top of the evaluation stack in the local variable list at index 1.

    :meta hide-value:
    """
    STLOC2 = b'\x72'
    """
    Stores the value on top of the evaluation stack in the local variable list at index 2.

    :meta hide-value:
    """
    STLOC3 = b'\x73'
    """
    Stores the value on top of the evaluation stack in the local variable list at index 3.

    :meta hide-value:
    """
    STLOC4 = b'\x74'
    """
    Stores the value on top of the evaluation stack in the local variable list at index 4.

    :meta hide-value:
    """
    STLOC5 = b'\x75'
    """
    Stores the value on top of the evaluation stack in the local variable list at index 5.

    :meta hide-value:
    """
    STLOC6 = b'\x76'
    """
    Stores the value on top of the evaluation stack in the local variable list at index 6.

    :meta hide-value:
    """
    STLOC = b'\x77'
    """
    Stores the value on top of the evaluation stack in the local variable list at a specified index. The index is
    represented as a 1-byte unsigned integer.

    :meta hide-value:
    """
    LDARG0 = b'\x78'
    """
    Loads the argument at index 0 onto the evaluation stack.

    :meta hide-value:
    """
    LDARG1 = b'\x79'
    """
    Loads the argument at index 1 onto the evaluation stack.

    :meta hide-value:
    """
    LDARG2 = b'\x7A'
    """
    Loads the argument at index 2 onto the evaluation stack.

    :meta hide-value:
    """
    LDARG3 = b'\x7B'
    """
    Loads the argument at index 3 onto the evaluation stack.

    :meta hide-value:
    """
    LDARG4 = b'\x7C'
    """
    Loads the argument at index 4 onto the evaluation stack.

    :meta hide-value:
    """
    LDARG5 = b'\x7D'
    """
    Loads the argument at index 5 onto the evaluation stack.

    :meta hide-value:
    """
    LDARG6 = b'\x7E'
    """
    Loads the argument at index 6 onto the evaluation stack.

    :meta hide-value:
    """
    LDARG = b'\x7F'
    """
    Loads the argument at a specified index onto the evaluation stack. The index is represented as a 1-byte
    unsigned integer.

    :meta hide-value:
    """
    STARG0 = b'\x80'
    """
    Stores the value on top of the evaluation stack in the argument slot at index 0.

    :meta hide-value:
    """
    STARG1 = b'\x81'
    """
    Stores the value on top of the evaluation stack in the argument slot at index 1.

    :meta hide-value:
    """
    STARG2 = b'\x82'
    """
    Stores the value on top of the evaluation stack in the argument slot at index 2.

    :meta hide-value:
    """
    STARG3 = b'\x83'
    """
    Stores the value on top of the evaluation stack in the argument slot at index 3.

    :meta hide-value:
    """
    STARG4 = b'\x84'
    """
    Stores the value on top of the evaluation stack in the argument slot at index 4.

    :meta hide-value:
    """
    STARG5 = b'\x85'
    """
    Stores the value on top of the evaluation stack in the argument slot at index 5.

    :meta hide-value:
    """
    STARG6 = b'\x86'
    """
    Stores the value on top of the evaluation stack in the argument slot at index 6.

    :meta hide-value:
    """
    STARG = b'\x87'
    """
    Stores the value on top of the evaluation stack in the argument slot at a specified index. The index is
    represented as a 1-byte unsigned integer.

    :meta hide-value:
    """

    # endregion

    # region Splice

    NEWBUFFER = b'\x88'
    """
    Creates a new Buffer and pushes it onto the stack.

    :meta hide-value:
    """
    MEMCPY = b'\x89'
    """
    Copies a range of bytes from one Buffer to another.

    :meta hide-value:
    """
    CAT = b'\x8B'
    """
    Concatenates two strings.

    :meta hide-value:
    """
    SUBSTR = b'\x8C'
    """
    Returns a section of a string.

    :meta hide-value:
    """
    LEFT = b'\x8D'
    """
    Keeps only characters left of the specified point in a string.

    :meta hide-value:
    """
    RIGHT = b'\x8E'
    """
    Keeps only characters right of the specified point in a string.

    :meta hide-value:
    """

    # endregion

    # region Bitwise logic

    INVERT = b'\x90'
    """
    Flips all of the bits in the input.

    :meta hide-value:
    """
    AND = b'\x91'
    """
    Boolean and between each bit in the inputs.

    :meta hide-value:
    """
    OR = b'\x92'
    """
    Boolean or between each bit in the inputs.

    :meta hide-value:
    """
    XOR = b'\x93'
    """
    Boolean exclusive or between each bit in the inputs.

    :meta hide-value:
    """
    EQUAL = b'\x97'
    """
    Returns 1 if the inputs are exactly equal, 0 otherwise.

    :meta hide-value:
    """
    NOTEQUAL = b'\x98'
    """
    Returns 1 if the inputs are not equal, 0 otherwise.

    :meta hide-value:
    """

    # endregion

    # region Arithmetic

    SIGN = b'\x99'
    """
    Puts the sign of top stack item on top of the main stack. If value is negative, put -1; if positive,
    put 1; if value is zero, put 0.

    :meta hide-value:
    """
    ABS = b'\x9A'
    """
    The input is made positive.

    :meta hide-value:
    """
    NEGATE = b'\x9B'
    """
    The sign of the input is flipped.

    :meta hide-value:
    """
    INC = b'\x9C'
    """
    1 is added to the input.

    :meta hide-value:
    """
    DEC = b'\x9D'
    """
    1 is subtracted from the input.

    :meta hide-value:
    """
    ADD = b'\x9E'
    """
    a is added to b.

    :meta hide-value:
    """
    SUB = b'\x9F'
    """
    b is subtracted from a.

    :meta hide-value:
    """
    MUL = b'\xA0'
    """
    a is multiplied by b.

    :meta hide-value:
    """
    DIV = b'\xA1'
    """
    a is divided by b.

    :meta hide-value:
    """
    MOD = b'\xA2'
    """
    Returns the remainder after dividing a by b.

    :meta hide-value:
    """
    POW = b'\xA3'
    """
    The result of raising value to the exponent power.

    :meta hide-value:
    """
    SQRT = b'\xA4'
    """
    Returns the square root of a specified number.

    :meta hide-value:
    """
    MODMUL = b'\xA5'
    """
    Performs modulus division on a number multiplied by another number.

    :meta hide-value:
    """
    MODPOW = b'\xA6'
    """
    Performs modulus division on a number raised to the power of another number.
    If the exponent is -1, it will have the calculation of the modular inverse.

    :meta hide-value:
    """
    SHL = b'\xA8'
    """
    Shifts a left b bits, preserving sign.

    :meta hide-value:
    """
    SHR = b'\xA9'
    """
    Shifts a right b bits, preserving sign.

    :meta hide-value:
    """
    NOT = b'\xAA'
    """
    If the input is 0 or 1, it is flipped. Otherwise the output will be 0.

    :meta hide-value:
    """
    BOOLAND = b'\xAB'
    """
    If both a and b are not 0, the output is 1. Otherwise 0.

    :meta hide-value:
    """
    BOOLOR = b'\xAC'
    """
    If a or b is not 0, the output is 1. Otherwise 0.

    :meta hide-value:
    """
    NZ = b'\xB1'
    """
    Returns 0 if the input is 0. 1 otherwise.

    :meta hide-value:
    """
    NUMEQUAL = b'\xB3'
    """
    Returns 1 if the numbers are equal, 0 otherwise.

    :meta hide-value:
    """
    NUMNOTEQUAL = b'\xB4'
    """
    Returns 1 if the numbers are not equal, 0 otherwise.

    :meta hide-value:
    """
    LT = b'\xB5'
    """
    Returns 1 if a is less than b, 0 otherwise.

    :meta hide-value:
    """
    LE = b'\xB6'
    """
    Returns 1 if a is less than or equal to b, 0 otherwise.

    :meta hide-value:
    """
    GT = b'\xB7'
    """
    Returns 1 if a is greater than b, 0 otherwise.

    :meta hide-value:
    """
    GE = b'\xB8'
    """
    Returns 1 if a is greater than or equal to b, 0 otherwise.

    :meta hide-value:
    """
    MIN = b'\xB9'
    """
    Returns the smaller of a and b.

    :meta hide-value:
    """
    MAX = b'\xBA'
    """
    Returns the larger of a and b.

    :meta hide-value:
    """
    WITHIN = b'\xBB'
    """
    Returns 1 if x is within the specified range (left-inclusive), 0 otherwise.

    :meta hide-value:
    """

    # endregion

    # region Compound-type

    PACKMAP = b'\xBE'
    """
    A value n is taken from top of main stack. The next n*2 items on main stack are removed, put inside n-sized map
    and this map is put on top of the main stack.

    :meta hide-value:
    """
    PACKSTRUCT = b'\xBF'
    """
    A value n is taken from top of main stack. The next n items on main stack are removed, put inside n-sized
    struct and this struct is put on top of the main stack.

    :meta hide-value:
    """
    PACK = b'\xC0'
    """
    A value n is taken from top of main stack. The next n items on main stack are removed, put inside n-sized array
    and this array is put on top of the main stack.

    :meta hide-value:
    """
    UNPACK = b'\xC1'
    """
    A collection is removed from top of the main stack. Its elements are put on top of the main stack (in reverse order) and the collection size is also put on main stack.

    :meta hide-value:
    """
    NEWARRAY0 = b'\xC2'
    """
    An empty array (with size 0) is put on top of the main stack.

    :meta hide-value:
    """
    NEWARRAY = b'\xC3'
    """
    A value n is taken from top of main stack. A null-filled array with size n is put on top of the main stack.

    :meta hide-value:
    """
    NEWARRAY_T = b'\xC4'
    """
    A value n is taken from top of main stack. An array of type T with size n is put on top of the main stack.

    :meta hide-value:
    """
    NEWSTRUCT0 = b'\xC5'
    """
    An empty struct (with size 0) is put on top of the main stack.

    :meta hide-value:
    """
    NEWSTRUCT = b'\xC6'
    """
    A value n is taken from top of main stack. A zero-filled struct with size n is put on top of the main stack.

    :meta hide-value:
    """
    NEWMAP = b'\xC8'
    """
    A Map is created and put on top of the main stack.

    :meta hide-value:
    """
    SIZE = b'\xCA'
    """
    An array is removed from top of the main stack. Its size is put on top of the main stack.

    :meta hide-value:
    """
    HASKEY = b'\xCB'
    """
    An input index n (or key) and an array (or map) are removed from the top of the main stack. Puts True on top of
    main stack if array[n] (or map[n]) exist, and False otherwise.

    :meta hide-value:
    """
    KEYS = b'\xCC'
    """
    A map is taken from top of the main stack. The keys of this map are put on top of the main stack.

    :meta hide-value:
    """
    VALUES = b'\xCD'
    """
    A map is taken from top of the main stack. The values of this map are put on top of the main stack.

    :meta hide-value:
    """
    PICKITEM = b'\xCE'
    """
    An input index n (or key) and an array (or map) are taken from main stack. Element array[n] (or map[n]) is put
    on top of the main stack.

    :meta hide-value:
    """
    APPEND = b'\xCF'
    """
    The item on top of main stack is removed and appended to the second item on top of the main stack.

    :meta hide-value:
    """
    SETITEM = b'\xD0'
    """
    A value v, index n (or key) and an array (or map) are taken from main stack. Attribution array[n]=v
    (or map[n]=v) is performed.

    :meta hide-value:
    """
    REVERSEITEMS = b'\xD1'
    """
    An array is removed from the top of the main stack and its elements are reversed.

    :meta hide-value:
    """
    REMOVE = b'\xD2'
    """
    An input index n (or key) and an array (or map) are removed from the top of the main stack. Element array[n]
    (or map[n]) is removed.

    :meta hide-value:
    """
    CLEARITEMS = b'\xD3'
    """
    Remove all the items from the compound-type.

    :meta hide-value:
    """
    POPITEM = b'\xD4'
    """
    Remove the last element from an array, and push it onto the stack.

    :meta hide-value:
    """

    # endregion

    # region Types

    ISNULL = b'\xD8'
    """
    Returns true if the input is null;

    :meta hide-value:
    """
    ISTYPE = b'\xD9'
    """
    Returns true if the top item of the stack is of the specified type;

    :meta hide-value:
    """
    CONVERT = b'\xDB'
    """
    Returns true if the input is null;

    :meta hide-value:
    """

    # endregion

    def __repr__(self) -> str:
        return str(self)



from typing import Any

from boa3.model.type.itype import IType


class BaseExceptionType(IType):
    """
    An class used to represent a generic Python exception
    """

    def __init__(self):
        identifier: str = 'exception'
        super().__init__(identifier)

    @property
    def is_generic(self) -> bool:
        return True

    @classmethod
    def build(cls, value: Any) -> IType:
        if cls._is_type_of(value):
            return cls()

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, BaseExceptionType)

    def __hash__(self):
        return hash(self.identifier)



import ast

from boa3.analyser.analyser import Analyser
from boa3.compiler.codegenerator.codegenerator import CodeGenerator
from boa3.model.type.type import Type
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest


class TestConstant(BoaTest):
    def build_code_generator(self) -> CodeGenerator:
        from boa3.compiler.codegenerator.vmcodemapping import VMCodeMapping
        VMCodeMapping.reset()
        return CodeGenerator({})

    def test_small_integer_constant(self):
        input = 7
        expected_output = Opcode.PUSH7

        generator = self.build_code_generator()
        generator.convert_integer_literal(input)
        output = generator.bytecode

        self.assertEqual(expected_output, output)

    def test_small_negative_integer_constant(self):
        input = -10
        expected_output = (
            Opcode.PUSH10
            + Opcode.NEGATE
        )

        generator = self.build_code_generator()
        generator.convert_integer_literal(input)
        output = generator.bytecode

        self.assertEqual(expected_output, output)

    def test_negative_integer_constant(self):
        input = -100
        byte_input = Integer(input).to_byte_array()
        expected_output = (
            Opcode.PUSHINT8             # push the bytes
            + byte_input
        )

        generator = self.build_code_generator()
        generator.convert_integer_literal(input)
        output = generator.bytecode

        self.assertEqual(expected_output, output)

    def test_one_byte_length_integer_constant(self):
        input = 42
        byte_input = Integer(input).to_byte_array()
        expected_output = (
            Opcode.PUSHINT8             # push the bytes
            + byte_input
        )

        generator = self.build_code_generator()
        generator.convert_integer_literal(input)
        output = generator.bytecode

        self.assertEqual(expected_output, output)

    def test_two_bytes_length_length_integer_constant(self):
        byte_input = bytes(300) + b'\x01'
        input = Integer.from_bytes(byte_input)
        expected_output = (
            Opcode.PUSHINT256           # push the bytes
            + byte_input[:32]
        )

        generator = self.build_code_generator()
        generator.convert_integer_literal(input)
        output = generator.bytecode

        self.assertEqual(expected_output, output)

    def test_big_integer_constant(self):
        byte_input = bytes(100000) + b'\x01'
        input = Integer.from_bytes(byte_input)
        expected_output = (
            Opcode.PUSHINT256           # push the bytes
            + byte_input[:32]
        )

        generator = self.build_code_generator()
        generator.convert_integer_literal(input)
        output = generator.bytecode

        self.assertEqual(expected_output, output)

    def test_ambiguous_integer_constant(self):
        byte_input = b'\x00\x80'
        unsigned = Integer.from_bytes(byte_input, signed=False)
        signed = Integer.from_bytes(byte_input, signed=True)
        self.assertNotEqual(unsigned, signed)

        unsigned_byte_input = Integer(unsigned).to_byte_array(signed=True, min_length=4)
        self.assertNotEqual(byte_input, unsigned_byte_input)
        unsigned_expected_output = (
            Opcode.PUSHINT32            # push the bytes
            + unsigned_byte_input
        )

        generator = self.build_code_generator()
        generator.convert_integer_literal(unsigned)
        output = generator.bytecode

        self.assertEqual(unsigned_expected_output, output)

        signed_byte_input = Integer(signed).to_byte_array(signed=True, min_length=2)
        self.assertEqual(byte_input, signed_byte_input)
        signed_expected_output = (
            Opcode.PUSHINT16           # push the bytes
            + signed_byte_input
        )

        generator = self.build_code_generator()
        generator.convert_integer_literal(signed)
        output = generator.bytecode

        self.assertEqual(signed_expected_output, output)

    def test_string_constant(self):
        input = 'unit_test'
        byte_input = String(input).to_bytes()
        expected_output = (
            Opcode.PUSHDATA1            # push the bytes
            + Integer(len(byte_input)).to_byte_array()
            + byte_input
        )

        generator = self.build_code_generator()
        generator.convert_string_literal(input)
        output = generator.bytecode

        self.assertEqual(expected_output, output)

    def test_integer_tuple_constant(self):
        input = (1, 2, 3)
        expected_output = (
            Opcode.PUSH3        # 3
            + Opcode.PUSH2      # 2
            + Opcode.PUSH1      # 1
            + Opcode.PUSH3      # tuple length
            + Opcode.PACK
        )

        generator = self.build_code_generator()
        generator.convert_literal(input)
        output = generator.bytecode

        self.assertEqual(expected_output, output)

    def test_string_tuple_constant(self):
        input = ('1', '2', '3')
        byte_input0 = String(input[0]).to_bytes()
        byte_input1 = String(input[1]).to_bytes()
        byte_input2 = String(input[2]).to_bytes()

        expected_output = (
            Opcode.INITSSLOT + b'\x01'
            + Opcode.PUSHDATA1  # '3'
            + Integer(len(byte_input2)).to_byte_array()
            + byte_input2
            + Opcode.PUSHDATA1  # '2'
            + Integer(len(byte_input1)).to_byte_array()
            + byte_input1
            + Opcode.PUSHDATA1  # '1'
            + Integer(len(byte_input0)).to_byte_array()
            + byte_input0
            + Opcode.PUSH3      # tuple length
            + Opcode.PACK
            + Opcode.DROP
            + Opcode.RET
        )

        analyser = Analyser(ast.parse(str(input)))
        analyser.symbol_table['x'] = Variable(Type.any)
        output = CodeGenerator.generate_code(analyser).bytecode

        self.assertEqual(expected_output, output)

    def test_any_tuple_constant(self):
        input = (1, '2', False)
        byte_input1 = String(input[1]).to_bytes()

        expected_output = (
            Opcode.PUSHF        # False
            + Opcode.PUSHDATA1  # '2'
            + Integer(len(byte_input1)).to_byte_array()
            + byte_input1
            + Opcode.PUSH1      # 1
            + Opcode.PUSH3      # tuple length
            + Opcode.PACK
        )

        generator = self.build_code_generator()
        generator.convert_literal(input)
        output = generator.bytecode

        self.assertEqual(expected_output, output)

    def test_tuple_of_tuple_constant(self):
        input = ((1, 2), (3, 4, 5, 6), (7,))
        expected_output = (
            # tuple[2]
            Opcode.PUSH7    # 7
            + Opcode.PUSH1  # tuple length
            + Opcode.PACK
            # tuple[1]
            + Opcode.PUSH6  # 6
            + Opcode.PUSH5  # 7
            + Opcode.PUSH4  # 4
            + Opcode.PUSH3  # 3
            + Opcode.PUSH4  # tuple length
            + Opcode.PACK
            # tuple[0]
            + Opcode.PUSH2  # 2
            + Opcode.PUSH1  # 1
            + Opcode.PUSH2  # tuple length
            + Opcode.PACK
            + Opcode.PUSH3  # tuple length
            + Opcode.PACK
        )

        generator = self.build_code_generator()
        generator.convert_literal(input)
        output = generator.bytecode

        self.assertEqual(expected_output, output)

    def test_integer_list_constant(self):
        input = [1, 2, 3]
        expected_output = (
            Opcode.PUSH3    # 3
            + Opcode.PUSH2  # 2
            + Opcode.PUSH1  # 1
            + Opcode.PUSH3  # list length
            + Opcode.PACK
        )

        generator = self.build_code_generator()
        generator.convert_literal(input)
        output = generator.bytecode

        self.assertEqual(expected_output, output)

    def test_string_list_constant(self):
        input = ['1', '2', '3']
        byte_input0 = String(input[0]).to_bytes()
        byte_input1 = String(input[1]).to_bytes()
        byte_input2 = String(input[2]).to_bytes()

        expected_output = (
            Opcode.PUSHDATA1        # '2'
            + Integer(len(byte_input2)).to_byte_array()
            + byte_input2
            + Opcode.PUSHDATA1      # '1'
            + Integer(len(byte_input1)).to_byte_array()
            + byte_input1
            + Opcode.PUSHDATA1      # '0'
            + Integer(len(byte_input0)).to_byte_array()
            + byte_input0
            + Opcode.PUSH3          # list length
            + Opcode.PACK
        )

        generator = self.build_code_generator()
        generator.convert_literal(input)
        output = generator.bytecode

        self.assertEqual(expected_output, output)

    def test_any_list_constant(self):
        input = [1, '2', False]
        byte_input1 = String(input[1]).to_bytes()

        expected_output = (
            Opcode.PUSHF        # False
            + Opcode.PUSHDATA1  # '2'
            + Integer(len(byte_input1)).to_byte_array()
            + byte_input1
            + Opcode.PUSH1      # 1
            + Opcode.PUSH3      # list length
            + Opcode.PACK
        )

        generator = self.build_code_generator()
        generator.convert_literal(input)
        output = generator.bytecode

        self.assertEqual(expected_output, output)

    def test_list_of_list_constant(self):
        input = [[1, 2], [3, 4, 5, 6], [7]]
        expected_output = (
            # list[2]
            Opcode.PUSH7    # 7
            + Opcode.PUSH1  # list length
            + Opcode.PACK
            # list[1]
            + Opcode.PUSH6  # 6
            + Opcode.PUSH5  # 5
            + Opcode.PUSH4  # 4
            + Opcode.PUSH3  # 3
            + Opcode.PUSH4  # list length
            + Opcode.PACK
            # list[0]
            + Opcode.PUSH2  # 2
            + Opcode.PUSH1  # 1
            + Opcode.PUSH2  # list length
            + Opcode.PACK
            + Opcode.PUSH3  # list length
            + Opcode.PACK
        )

        generator = self.build_code_generator()
        generator.convert_literal(input)
        output = generator.bytecode

        self.assertEqual(expected_output, output)

    def test_integer_dict_constant(self):
        input = {1: 15, 2: 14, 3: 13}
        expected_output = (
            Opcode.NEWMAP     # {1: 15, 2: 14, 3: 13}
            + Opcode.DUP
            + Opcode.PUSH1      # map[1] = 15
            + Opcode.PUSH15
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH2      # map[2] = 14
            + Opcode.PUSH14
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH3      # map[3] = 13
            + Opcode.PUSH13
            + Opcode.SETITEM
        )

        generator = self.build_code_generator()
        generator.convert_literal(input)
        output = generator.bytecode

        self.assertEqual(expected_output, output)

    def test_string_dict_constant(self):
        input = {'one': 1, 'two': 2, 'three': 3}
        one = String('one').to_bytes()
        two = String('two').to_bytes()
        three = String('three').to_bytes()

        expected_output = (
            Opcode.NEWMAP     # {'one': 1, 'two': 2, 'three': 3}
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['one'] = 1
            + Integer(len(one)).to_byte_array(min_length=1)
            + one
            + Opcode.PUSH1
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['two'] = 2
            + Integer(len(two)).to_byte_array(min_length=1)
            + two
            + Opcode.PUSH2
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSHDATA1  # map['three'] = 3
            + Integer(len(three)).to_byte_array(min_length=1)
            + three
            + Opcode.PUSH3
            + Opcode.SETITEM
        )

        generator = self.build_code_generator()
        generator.convert_literal(input)
        output = generator.bytecode

        self.assertEqual(expected_output, output)

    def test_any_dict_constant(self):
        input = {1: True, 2: 4, 3: 'nine'}
        nine = String('nine').to_bytes()

        expected_output = (
            Opcode.NEWMAP  # {1: True, 2: 4, 3: 'nine'}
            + Opcode.DUP
            + Opcode.PUSH1  # map[1] = True
            + Opcode.PUSHT
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH2  # map[2] = 4
            + Opcode.PUSH4
            + Opcode.SETITEM
            + Opcode.DUP
            + Opcode.PUSH3  # map[3] = 'nine'
            + Opcode.PUSHDATA1
            + Integer(len(nine)).to_byte_array(min_length=1)
            + nine
            + Opcode.SETITEM
        )

        generator = self.build_code_generator()
        generator.convert_literal(input)
        output = generator.bytecode

        self.assertEqual(expected_output, output)

    def test_integer_script_hash(self):
        from boa3.neo import cryptography, to_script_hash

        input = Integer(123).to_byte_array()
        expected_output = cryptography.hash160(input)
        output = to_script_hash(input)

        self.assertEqual(expected_output, output)

    def test_string_script_hash(self):
        import base58
        from boa3.neo import to_script_hash

        input = String('123').to_bytes()
        expected_output = base58.b58decode(input)[1:]
        output = to_script_hash(input)

        self.assertEqual(expected_output, output)

    def test_bytes_script_hash(self):
        from boa3.neo import cryptography, to_script_hash

        input = b'\x01\x02\x03'
        expected_output = cryptography.hash160(input)
        output = to_script_hash(input)

        self.assertEqual(expected_output, output)

    def test_address_script_hash(self):
        from base58 import b58decode
        from boa3.neo import cryptography, to_script_hash

        input = String('Nd7eAuHsKvvzHzSPyuJQALcYCcUrcwvm5W').to_bytes()
        expected_output = b58decode(input)[1:21]
        wrong_output = cryptography.hash160(input)
        output = to_script_hash(input)

        self.assertNotEqual(wrong_output, output)
        self.assertEqual(expected_output, output)



from typing import List, Optional, Tuple

from boa3.model.builtin.classmethod.copymethod import CopyMethod
from boa3.model.type.itype import IType
from boa3.neo.vm.opcode import OpcodeHelper
from boa3.neo.vm.opcode.Opcode import Opcode


class CopyDictMethod(CopyMethod):

    def __init__(self, arg_value: Optional[IType] = None):
        from boa3.model.type.type import Type
        super().__init__(arg_value if Type.dict.is_type_of(arg_value) else Type.dict)

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.compiler.codegenerator import get_bytes_count

        jmp_place_holder = (Opcode.JMP, b'\x01')

        create_auxiliary_vars = [       # initializes auxiliary variables
            (Opcode.NEWMAP, b''),       # dict_copy = {}
            (Opcode.OVER, b''),
            (Opcode.KEYS, b''),         # list_keys = list(dict_target)
            (Opcode.DUP, b''),
            (Opcode.SIZE, b''),
            (Opcode.DEC, b''),          # index = len(list_keys) - 1
        ]

        verify_while_condition = [      # if index < 0: go clean stack
            (Opcode.DUP, b''),
            (Opcode.PUSH0, b''),
            jmp_place_holder            # jump to clean_stack
        ]

        while_loop = [                  # while index >= 0: add key value pair into new dict
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.PICKITEM, b''),     # key = list_keys[index]
            (Opcode.PUSH5, b''),
            (Opcode.PICK, b''),
            (Opcode.OVER, b''),
            (Opcode.PICKITEM, b''),     # value = dict_target[key]
            (Opcode.SETITEM, b''),      # dict_copy[key] = value
            (Opcode.DEC, b''),          # index--
            # jump back to verify_while_condition
        ]

        jmp_back_to_verify = OpcodeHelper.get_jump_and_data(Opcode.JMP, -get_bytes_count(while_loop + verify_while_condition))
        while_loop.append(jmp_back_to_verify)

        jmp_while_loop = OpcodeHelper.get_jump_and_data(Opcode.JMPLT, get_bytes_count(while_loop))
        verify_while_condition[-1] = jmp_while_loop

        clean_stack = [                 # remove all values from stack except the dict_copy
            (Opcode.DROP, b''),
            (Opcode.DROP, b''),
            (Opcode.NIP, b'')
        ]

        return (
            create_auxiliary_vars +
            verify_while_condition +
            while_loop +
            clean_stack
        )



from typing import Dict, Optional

from boa3.model.builtin.builtinproperty import IBuiltinProperty
from boa3.model.builtin.interop.contractgethashmethod import ContractGetHashMethod
from boa3.model.type.collection.sequence.uint160type import UInt160Type
from boa3.model.variable import Variable


class ContractInterfaceGetScriptHashMethod(ContractGetHashMethod):
    def __init__(self, identifier: str, contract_script: bytes):
        identifier = f'-get_{identifier}'
        args: Dict[str, Variable] = {}
        super().__init__(contract_script, identifier, args, return_type=UInt160Type.build())

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None


class ContractHashProperty(IBuiltinProperty):
    def __init__(self, identifier: str, contract_script: bytes):
        getter = ContractInterfaceGetScriptHashMethod(identifier, contract_script)
        super().__init__(identifier, getter)



import ast
from typing import Dict, List, Optional, Tuple

from boa3.model.callable import Callable
from boa3.model.identifiedsymbol import IdentifiedSymbol
from boa3.model.type.type import Type
from boa3.model.variable import Variable


class Event(Callable, IdentifiedSymbol):
    def __init__(self, event_id: str, args: Dict[str, Variable] = None,
                 vararg: Optional[Tuple[str, Variable]] = None,
                 kwargs: Optional[Dict[str, Variable]] = None,
                 defaults: List[ast.AST] = None,
                 origin_node: Optional[ast.AST] = None):
        super().__init__(args, vararg, kwargs, defaults, Type.none, True, origin_node)

        self.name: str = event_id
        self._identifier: str = None

    @property
    def shadowing_name(self) -> str:
        return 'event'

    @property
    def identifier(self) -> str:
        if self._identifier is None:
            # internal identifier should not be the name to avoid nonexistent duplicated symbol ids
            self._identifier = f'-{self.name}-{hex(id(self))}'
        return self._identifier

    @property
    def args_to_generate(self) -> Dict[str, Variable]:
        return self.args.copy()

    @property
    def generate_name(self) -> bool:
        return True



from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestContractInterface(BoaTest):
    default_folder: str = 'test_sc/contract_interface_test'

    def test_contract_interface_decorator_literal_hash_str(self):
        path = self.get_contract_path('ContractInterfaceLiteralStrHash.py')
        Boa3.compile(path)  # test if compiles because the smart contract doesn't exist

    def test_contract_interface_decorator_literal_hash_bytes(self):
        path = self.get_contract_path('ContractInterfaceLiteralBytesHash.py')
        Boa3.compile(path)  # test if compiles because the smart contract doesn't exist

    def test_contract_interface_decorator_invalid_hash(self):
        path = self.get_contract_path('ContractInterfaceInvalidHash.py')
        self.assertCompilerLogs(CompilerError.InvalidUsage, path)

    def test_contract_interface_decorator_variable_hash(self):
        path = self.get_contract_path('ContractInterfaceVariableArgument.py')
        self.assertCompilerLogs(CompilerError.InvalidUsage, path)

    def test_contract_interface_decorator_too_few_arguments(self):
        path = self.get_contract_path('ContractInterfaceTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_contract_interface_decorator_too_many_arguments(self):
        path = self.get_contract_path('ContractInterfaceTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_contract_interface_decorator_without_call(self):
        path = self.get_contract_path('ContractInterfaceWithoutCall.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_contract_interface_with_instance_method(self):
        path = self.get_contract_path('ContractInterfaceInstanceMethod.py')
        self.assertCompilerLogs(CompilerError.InvalidUsage, path)

    def test_contract_interface_with_class_method(self):
        path = self.get_contract_path('ContractInterfaceClassMethod.py')
        self.assertCompilerLogs(CompilerError.InvalidUsage, path)

    def test_contract_interface_nep17(self):
        path = self.get_contract_path('Nep17Interface.py')
        nep17_path = self.get_contract_path('examples', 'nep17.py')

        engine = TestEngine()

        nep17_result = self.run_smart_contract(engine, nep17_path, 'symbol')
        result = self.run_smart_contract(engine, path, 'nep17_symbol')
        self.assertEqual(nep17_result, result)

    def test_contract_interface_display_name_argument(self):
        path = self.get_contract_path('ContractInterfaceDisplayNameRegularArgument.py')
        Boa3.compile(path)  # test if compiles because the smart contract doesn't exist

    def test_contract_interface_display_name_keyword_argument(self):
        path = self.get_contract_path('ContractInterfaceDisplayNameKeywordArgument.py')
        Boa3.compile(path)  # test if compiles because the smart contract doesn't exist

    def test_contract_interface_display_name_variable_name(self):
        path = self.get_contract_path('ContractInterfaceDisplayNameVariableArgument.py')
        self.assertCompilerLogs(CompilerError.InvalidUsage, path)

    def test_contract_interface_display_name_too_few_arguments(self):
        path = self.get_contract_path('ContractInterfaceDisplayNameTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_contract_interface_display_name_without_call(self):
        path = self.get_contract_path('ContractInterfaceDisplayNameWithoutCall.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_contract_interface_display_name_too_many_arguments(self):
        path = self.get_contract_path('ContractInterfaceDisplayNameTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_contract_interface_nep17_with_display_name(self):
        path = self.get_contract_path('Nep17InterfaceWithDisplayName.py')
        nep17_path = self.get_contract_path('examples', 'nep17.py')

        engine = TestEngine()

        nep17_result = self.run_smart_contract(engine, nep17_path, 'totalSupply')
        result = self.run_smart_contract(engine, path, 'nep17_total_supply')
        self.assertEqual(nep17_result, result)

    def test_contract_interface_code_optimization(self):
        from boa3.model.builtin.interop.interop import Interop
        from boa3.neo.vm.opcode.Opcode import Opcode
        from boa3.neo.vm.type.Integer import Integer
        from boa3.neo.vm.type.String import String
        from boa3.neo3.core.types import UInt160

        external_contract_name = 'symbol'
        function_name_bytes = String(external_contract_name).to_bytes()
        contract_script_bytes = UInt160.from_string('cc370e7538a8729221e6eb444eb6f8838d256042').to_array()

        expected_output = (
            Opcode.NEWARRAY0    # arguments list
            + Opcode.PUSH15     # CallFlag
            + Opcode.PUSHDATA1  # function name
            + Integer(len(function_name_bytes)).to_byte_array()
            + function_name_bytes
            + Opcode.PUSHDATA1  # contract script
            + Integer(len(contract_script_bytes)).to_byte_array()
            + contract_script_bytes
            + Opcode.SYSCALL
            + Interop.CallContract.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('ContractInterfaceCodeOptimization.py')
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        nep17_path = self.get_contract_path('examples', 'nep17.py')
        engine = TestEngine()

        nep17_result = self.run_smart_contract(engine, nep17_path, 'symbol')
        result = self.run_smart_contract(engine, path, 'nep17_symbol')
        self.assertEqual(nep17_result, result)

    def test_contract_manual_interface_code_optimization(self):
        from boa3.model.builtin.interop.interop import Interop
        from boa3.neo.vm.opcode.Opcode import Opcode
        from boa3.neo.vm.type.Integer import Integer
        from boa3.neo.vm.type.String import String
        from boa3.neo3.core.types import UInt160

        external_contract_name = 'symbol'
        function_name_bytes = String(external_contract_name).to_bytes()
        contract_script_bytes = UInt160.from_string('cc370e7538a8729221e6eb444eb6f8838d256042').to_array()

        expected_output = (
            # start public method
            Opcode.LDSFLD0      # generated cls arg
            + Opcode.CALL
            + Integer(35).to_byte_array()
            + Opcode.RET
            # end public method
            # start initialize method
            + Opcode.INITSSLOT + b'\x01'
            + Opcode.PUSH1
            + Opcode.NEWARRAY
            + Opcode.STSFLD0
            + Opcode.PUSHDATA1
            + Integer(len(contract_script_bytes)).to_byte_array()
            + contract_script_bytes
            + Opcode.PUSH0
            + Opcode.LDSFLD0
            + Opcode.REVERSE3
            + Opcode.SETITEM
            + Opcode.RET
            # end initialize method
            # start 'symbol' class method
            + Opcode.INITSLOT + b'\x00\x01'
            + Opcode.NEWARRAY0    # arguments list
            + Opcode.PUSH15     # CallFlag
            + Opcode.PUSHDATA1  # function name
            + Integer(len(function_name_bytes)).to_byte_array()
            + function_name_bytes
            + Opcode.LDARG0     # contract script
            + Opcode.PUSH0
            + Opcode.PICKITEM
            + Opcode.SYSCALL
            + Interop.CallContract.interop_method_hash
            + Opcode.RET
            # start class method
        )

        path = self.get_contract_path('ContractManualInterfaceCodeOptimization.py')
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        nep17_path = self.get_contract_path('examples', 'nep17.py')
        engine = TestEngine()

        nep17_result = self.run_smart_contract(engine, nep17_path, 'symbol')
        result = self.run_smart_contract(engine, path, 'nep17_symbol')
        self.assertEqual(nep17_result, result)

    def test_get_hash(self):
        path = self.get_contract_path('ContractInterfaceGetHash.py')
        engine = TestEngine()

        contract_script_bytes = bytes(reversed(range(20)))
        result = self.run_smart_contract(engine, path, 'main',
                                         expected_result_type=bytes)
        self.assertEqual(contract_script_bytes, result)



from boa3.builtin.compile_time import public
from boa3_test.test_sc.import_test.FromImportUserModuleRecursiveImport import from_import_empty_list


@public
def empty_list() -> list:
    return from_import_empty_list()
